from attr import attrib, NOTHING
from related import _init_fields, types
from collections import OrderedDict
from .converters import to_sequence_field_w_str, to_leaf_mapping_field, to_eval_str, identity
from . import dispatchers  # to load the dispatcher


class UNSPECIFIED(object):
    pass


def AnyField(default=NOTHING, required=True, repr=True):
    """
    Just pass through the field, using default yaml conversion to python objects

    :param cls: class (or name) of the model to be related in Sequence.
    :param default: any TypedSequence or list
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    """
    default = _init_fields.init_default(required, default, UNSPECIFIED())
    return attrib(default=default, convert=None, validator=None,
                  repr=repr)


def StrSequenceField(cls, default=NOTHING, required=True, repr=True):
    """
    Create new sequence field on a model. If only string is present,
    convert it to a list of length 1.

    :param cls: class (or name) of the model to be related in Sequence.
    :param default: any TypedSequence or list
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    """
    default = _init_fields.init_default(required, default, [])
    # check that it's not sequence
    converter = to_sequence_field_w_str(cls)
    validator = _init_fields.init_validator(required, types.TypedSequence)
    return attrib(default=default, convert=converter, validator=validator,
                  repr=repr)


def NestedMappingField(cls, keyword, key, default=NOTHING, required=True, repr=False):
    """
    Create new sequence field on a model. If only string is present,
    convert it to a list of length 1.

    :param cls: class (or name) of the model to be related in Sequence.
    :param keyword: stopping condition in recursion (indicator that cls has been found)
    :param key: key field on the child object to be used as the mapping key.
    :param default: any TypedSequence or list
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    """
    default = _init_fields.init_default(required, default, OrderedDict())
    # check that it's not sequence
    converter = to_leaf_mapping_field(cls, keyword, key)
    # validator = _init_fields.init_validator(required, types.TypedSequence)
    validator = None
    return attrib(default=default, convert=converter, validator=validator,
                  repr=repr)


def TupleIntField(default=NOTHING, required=True, repr=True):
    """
    Create new tuple field on a model. Convert it first to a string
    and then to a tuple

    :param cls: class (or name) of the model to be related in Sequence.
    :param default: any TypedSequence or list
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    """
    default = _init_fields.init_default(required, default, tuple)
    converter = to_eval_str
    validator = _init_fields.init_validator(required, tuple)
    return attrib(default=default, convert=converter, validator=validator,
                  repr=repr)



"""Conda helper files

Reusing code from: https://github.com/conda/conda-api/blob/master/conda_api.py
"""
from __future__ import absolute_import
from __future__ import print_function

import os
import json
import sys
import subprocess
from subprocess import Popen, PIPE, STDOUT
from collections import OrderedDict
from kipoi.utils import yaml_ordered_dump, unique_list
import six
import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class CondaError(Exception):
    "General Conda error"
    pass


def create_env(env_name, conda_deps):
    """Create new environment given parsed dependencies

    Args:
      conda_dependencies: OrderedDict of the `dependencies` field in Conda's environment.yaml.
        `model.dependencies.conda` field in Model or Dataloader.
      env_name: Environment name
    """
    # check if the environment already exists
    if env_exists(env_name):
        logger.info("Conda environment: {0} already exists. Skipping the installation.".
                    format(env_name))
        return None

    # write the env to file
    env_dict = OrderedDict([
        ("name", env_name),
        ("dependencies", conda_deps)
    ])
    tmp_dir = "/tmp/kipoi"
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
    tmp_file_path = "{tmp_dir}/{env_name}.yml".format(tmp_dir=tmp_dir,
                                                      env_name=env_name)
    with open(tmp_file_path, 'w') as f:
        f.write(yaml_ordered_dump(env_dict, indent=4, default_flow_style=False))

    # create the environment
    return create_env_from_file(tmp_file_path)


def get_env_path(env_name):
    for env in get_envs():
        if env.endswith(env_name):
            return env
    return None


def get_kipoi_bin(env_name):
    """Returns the path to the kipoi executable in {env_name} environment
    """
    env_root = get_env_path(env_name)
    if env_root is None:
        raise ValueError("Conda environment {0} doesn't exist".format(env_name))
    out_path = os.path.join(env_root, "bin", "kipoi")
    if not os.path.exists(out_path):
        raise ValueError("kipoi is not installed in conda environment: {0}".format(env_name))
    return out_path


def create_env_from_file(env_file):
    cmd_list = ["env", "create", "--file", env_file]
    return _call_conda(cmd_list, use_stdout=True)


def install_conda(conda_deps, channels=["defaults"]):
    """Install conda packages

    Args:
      conda_deps: list of conda packages to be installed
      channels: list of conda channels to use
    """
    conda_deps_wo_python = [x for x in conda_deps if "python" != x[:6]]
    if conda_deps_wo_python:
        cmd_list = ["install", "-y"]
        if channels:
            cmd_list += ["--channel={0}".format(c) for c in channels] + ["--override-channels"]
            # `--override-channels` is here in order to increase reproducibility
            # on different computers with different ~/.condarc file
        cmd_list += conda_deps_wo_python
        return _call_conda(cmd_list, use_stdout=True)


def install_pip(pip_deps):
    if pip_deps:
        cmd_list = ["install"] + list(pip_deps)
        return _call_pip(cmd_list, use_stdout=True)


def remove_env(env_name):
    cmd_list = ["env", "remove", "-y", "-n", env_name]
    return _call_conda(cmd_list, use_stdout=True)


def get_envs():
    """
    Return all of the (named) environment (this does not include the root
    environment), as a list of absolute path to their prefixes.
    """
    info = _call_and_parse(['info', '--json'])
    return info['envs']


def env_exists(env):
    return env in [os.path.basename(x) for x in get_envs()]


def _call_command(cmd, extra_args, use_stdout=False,
                  return_logs_with_stdout=False, **kwargs):
    """
    Args:
      return_logs_with_stdout (bool): If True, return also the logged lines
          (it only takes an effect with use_stdout)
    """
    # call conda with the list of extra arguments, and return the tuple
    # stdout, stderr
    cmd_list = [cmd]  # just use whatever conda is on the path

    cmd_list.extend(extra_args)

    try:
        if use_stdout:
            p = Popen(cmd_list, stdout=PIPE, universal_newlines=True, **kwargs)
            # Poll process for new output until finished
            if return_logs_with_stdout:
                out = []
            for stdout_line in iter(p.stdout.readline, ""):
                print(stdout_line, end='')
                if return_logs_with_stdout:
                    out.append(stdout_line.rstrip())
            p.stdout.close()
            return_code = p.wait()
            if return_code:
                raise subprocess.CalledProcessError(return_code, cmd_list)
            if return_logs_with_stdout:
                return return_code, out
            else:
                return return_code
        else:
            p = Popen(cmd_list, stdout=PIPE, stderr=PIPE, **kwargs)
    except OSError:
        raise Exception("could not invoke {0}\n".format(cmd_list))
    return p.communicate()


def _call_conda(extra_args, use_stdout=False):
    return _call_command("conda", extra_args, use_stdout)


def _call_pip(extra_args, use_stdout=False):
    return _call_command("pip", extra_args, use_stdout)


def _call_and_parse(extra_args):
    stdout, stderr = _call_conda(extra_args)
    if stderr.decode().strip():
        raise Exception('conda %r:\nSTDERR:\n%s\nEND' % (extra_args,
                                                         stderr.decode()))
    return json.loads(stdout.decode())


def parse_conda_package(dep):
    """Parse conda package into channel and environment

    Args:
      dep: string in the form '<channel>::<package>' or '<package>'

    Returns:
      tuple: ('<channel>', '<package>') or ('defaults', '<package>')
    """
    if "::" in dep:
        try:
            channel, package = dep.split("::")
        except ValueError:
            raise ValueError("The conda dependency: {0} couldn't be properly parsed. ".format(dep) +
                             "Use the following synthax: <channel>::<package> or <package>")
        return (channel, package)
    else:
        return ("defaults", dep)


def version_split(s, delimiters={"=", ">", "<", "~"}):
    """Split the string by the version:
    mypacakge<=2.4,==2.4 -> (mypacakge, <=2.4,==2.4)

    In [40]: version_split("asdsda>=2.4,==2")
    Out[40]: ('asdsda', ['>=2.4', '==2'])

    In [41]: version_split("asdsda>=2.4")
    Out[41]: ('asdsda', ['>=2.4'])

    In [42]: version_split("asdsda")
    Out[42]: ('asdsda', [])
    """
    for i, c in enumerate(s):
        if c in delimiters:
            return (s[:i], s[i:].split(","))
    return (s, [])


def normalize_pip(pip_list):
    """Normalize a list of pip dependencies

    Args:
      pip_list: list of pip dependencies

    Returns:
      normalized pip
    """

    d_list = OrderedDict()
    for d in pip_list:
        package, versions = version_split(d)
        if package in d_list:
            d_list[package] = unique_list(d_list[package] + versions)
        else:
            d_list[package] = versions
    return [package + ",".join(versions) for package, versions in six.iteritems(d_list)]


def get_package_version(package):
    """Get installed package version

    # Arguments
      package: package name. Example: `kipoi`

    # Returns
      (str) if the package is installed and None otherwise
    """
    import pkg_resources
    try:
        return pkg_resources.get_distribution(package).version
    except pkg_resources.DistributionNotFound:
        return None


def compatible_versions(installed_version, req_version):
    from pkg_resources import parse_version
    installed_version = parse_version(installed_version)
    if req_version.startswith(">="):
        return installed_version >= parse_version(req_version[2:])
    elif req_version.startswith("<="):
        return installed_version <= parse_version(req_version[2:])
    elif req_version.startswith("=="):
        return installed_version == parse_version(req_version[2:])
    elif req_version.startswith("<"):
        return installed_version < parse_version(req_version[1:])
    elif req_version.startswith(">"):
        return installed_version > parse_version(req_version[1:])
    else:
        raise ValueError("Package prefix {} needs to be from >=,<=, ==, >, <".
                         format(req_version))


def is_installed(package):
    """Test if the package is installed

    # Arguments
      package: package string, can also contain version: mypacakge<=2.4,==2.4
    """
    package, required_versions = version_split(package)
    installed_version = get_package_version(package)
    if installed_version is None:
        # package is for sure not installed
        return False
    else:
        for req_version in required_versions:
            return compatible_versions(installed_version, req_version)
        return True



"""Dataloader
"""
import pickle
{%- if cookiecutter.dataloader_type not in ["PreloadedDataset", "SampleGenerator", "BatchGenerator"] %}
from kipoi.data import {{ cookiecutter.dataloader_type }}
{%- endif %}
import pandas as pd
import numpy as np
import os

# access the absolute path to this script
# https://stackoverflow.com/questions/3718657/how-to-properly-determine-current-script-directory-in-python
import inspect
this_file_path = os.path.abspath(inspect.getframeinfo(inspect.currentframe()).filename)
this_dir = os.path.dirname(this_file_path)


def read_pickle(f):
    with open(f, "rb") as f:
        return pickle.load(f)
{% macro return_dict(w_targets=True, ret='return') -%}
{{ ret }} {
    {%- if cookiecutter.model_input_type == 'np.array' %}
    "inputs": x_features,
    {%- elif cookiecutter.model_input_type == 'list of np.arrays' %}
    "inputs": [x_features],
    {%- elif cookiecutter.model_input_type == 'dict of np.arrays' %}
    "inputs": {
        "features": x_features  
    },
    {%- endif %}
    {%- if w_targets %}
    {%- if cookiecutter.model_output_type == 'np.array' %}
    "targets": y_class,
    {%- elif cookiecutter.model_output_type == 'list of np.arrays' %}
    "targets": [y_class],
    {%- elif cookiecutter.model_output_type == 'dict of np.arrays' %}
    "targets": {
        "iris_class": y_class
    },
    {%- endif %}
    {%- endif %}
    "metadata": {
        "example_row_number": idx
    }
}    
{%- endmacro %}
{% if cookiecutter.dataloader_type == 'Dataset' %}
class My{{ cookiecutter.dataloader_type }}(Dataset):

    def __init__(self, features_file, targets_file=None):
        self.features_file = features_file
        self.targets_file = targets_file

        self.x_transformer = read_pickle(this_dir + "/dataloader_files/x_transformer.pkl")

        self.features = pd.read_csv(features_file)
        if targets_file is not None:
            self.y_transformer = read_pickle(this_dir + "/dataloader_files/y_transformer.pkl")
            self.targets = pd.read_csv(targets_file)
            assert len(self.targets) == len(self.features)

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        x_features = np.ravel(self.x_transformer.transform(self.features.iloc[idx].values[np.newaxis]))
        if self.targets_file is None:
            {{return_dict(False)|indent(12)}}
        else:
            y_class = np.ravel(self.y_transformer.transform(self.targets.iloc[idx].values[np.newaxis]))
            {{return_dict(True)|indent(12)}}

{% elif cookiecutter.dataloader_type == 'PreloadedDataset' %}
def my{{ cookiecutter.dataloader_type }}(features_file, targets_file=None):
    x_transformer = read_pickle(this_dir + "/dataloader_files/x_transformer.pkl")

    x_features = x_transformer.transform(pd.read_csv(features_file).values)
    idx = np.arange(x_features.shape[0])
    if targets_file is None:
        {{return_dict(False)|indent(8)}}
    else:
        y_transformer = read_pickle(this_dir + "/dataloader_files/y_transformer.pkl")
        y_class = y_transformer.transform(pd.read_csv(targets_file).values)
        {{return_dict(True)|indent(8)}}

{% elif cookiecutter.dataloader_type == 'BatchDataset' %}
class My{{ cookiecutter.dataloader_type }}(BatchDataset):

    def __init__(self, features_file, batch_size=5, targets_file=None):
        self.features_file = features_file
        self.targets_file = targets_file
        self.batch_size = batch_size

        self.x_transformer = read_pickle(this_dir + "/dataloader_files/x_transformer.pkl")

        self.features = pd.read_csv(features_file)
        if targets_file is not None:
            self.y_transformer = read_pickle(this_dir + "/dataloader_files/y_transformer.pkl")
            self.targets = pd.read_csv(targets_file)

    def __len__(self):
        return int(np.ceil(len(self.features) / self.batch_size))

    def __getitem__(self, idx):
        start = idx * self.batch_size
        end = min(len(self.features), (idx + 1) * self.batch_size)
        x_features = self.x_transformer.transform(self.features.iloc[start:end].values)
        if self.targets_file is None:
            {{return_dict(False)|indent(12)}}
        else:
            y_class = self.y_transformer.transform(self.targets.iloc[start:end].values)
            {{return_dict(True)|indent(12)}}


{% elif cookiecutter.dataloader_type == 'SampleIterator' %}
def row2np(row):
    return np.array(row.strip().split(",")).astype(float)

class My{{ cookiecutter.dataloader_type }}(SampleIterator):

    def __init__(self, features_file, targets_file=None):
        self.features_file = features_file
        self.targets_file = targets_file

        self.x_transformer = read_pickle(this_dir + "/dataloader_files/x_transformer.pkl")

        self.features_it = open(self.features_file, "r")
        next(self.features_it)  # skip the header

        self.idx = 0
        if targets_file is not None:
            self.y_transformer = read_pickle(this_dir + "/dataloader_files/y_transformer.pkl")
            self.targets_it = open(self.targets_file, "r")
            next(self.targets_it)  # skip the header

    def __iter__(self):
        return self

    def __next__(self):
        x_features = np.ravel(self.x_transformer.transform(row2np(next(self.features_it))[np.newaxis]))
        idx = self.idx
        self.idx += 1
        if self.targets_file is None:
            {{return_dict(False)|indent(12)}}
        else:
            y_class = np.ravel(self.y_transformer.transform(row2np(next(self.targets_it))[np.newaxis]))
            {{return_dict(True)|indent(12)}}

    # python2
    next = __next__


{% elif cookiecutter.dataloader_type == 'BatchIterator' %}
def row2np(row):
    return np.array(row.strip().split(",")).astype(float)


class My{{ cookiecutter.dataloader_type }}(BatchIterator):

    def __init__(self, features_file, batch_size=5, targets_file=None):
        self.features_file = features_file
        self.targets_file = targets_file
        self.batch_size = batch_size

        self.x_transformer = read_pickle(this_dir + "/dataloader_files/x_transformer.pkl")

        self.features_it = open(self.features_file, "r")
        next(self.features_it)  # skip the header
        self.idx = 0

        if targets_file is not None:
            self.y_transformer = read_pickle(this_dir + "/dataloader_files/y_transformer.pkl")
            self.targets_it = open(self.targets_file, "r")
            next(self.targets_it)  # skip the header


    def __iter__(self):
        return self

    def __next__(self):
        x_feat = []
        y_targets = []
        idx = []
        start = self.batch_size * self.idx
        for i in range(self.batch_size):
            try:
                x_feat.append(row2np(next(self.features_it)))
                if self.features_file is not None:
                    y_targets.append(row2np(next(self.targets_it)))
                self.idx += 1
                idx.append(self.idx)
            except StopIteration:
                if len(x_feat) == 0:
                    raise StopIteration
                else:
                    break
        idx = np.array(idx)
        x_features = self.x_transformer.transform(np.stack(x_feat))
        if self.targets_file is None:
            {{return_dict(False)|indent(12)}}
        else:
            y_class = self.y_transformer.transform(np.stack(y_targets))
            {{return_dict(True)|indent(12)}}

    # python2
    next = __next__


{% elif cookiecutter.dataloader_type == 'SampleGenerator' %}
def row2np(row):
    return np.array(row.strip().split(",")).astype(float)


def my{{ cookiecutter.dataloader_type }}(features_file, targets_file=None):
    x_transformer = read_pickle(this_dir + "/dataloader_files/x_transformer.pkl")

    features_it = open(features_file, "r")
    next(features_it)  # skip the header
    if targets_file is not None:
        targets_it = open(targets_file, "r")
        y_transformer = read_pickle(this_dir + "/dataloader_files/y_transformer.pkl")
        next(targets_it)  # skip the header
    for idx, row in enumerate(features_it):
        x_features = np.ravel(x_transformer.transform(row2np(row)[np.newaxis]))
        if targets_file is None:
            {{return_dict(False, "yield")|indent(12)}}
        else:
            y_class = np.ravel(y_transformer.transform(row2np(next(targets_it))[np.newaxis]))
            {{return_dict(True, "yield")|indent(12)}}


{% elif cookiecutter.dataloader_type == 'BatchGenerator' %}
def row2np(row):
    return np.array(row.strip().split(",")).astype(float)

def my{{ cookiecutter.dataloader_type }}(features_file, targets_file=None, batch_size=5):
    x_transformer = read_pickle(this_dir + "/dataloader_files/x_transformer.pkl")

    features_it = open(features_file, "r")
    next(features_it)  # skip the header
    if targets_file is not None:
        targets_it = open(targets_file, "r")
        y_transformer = read_pickle(this_dir + "/dataloader_files/y_transformer.pkl")
        next(targets_it)  # skip the header
    gidx = 0
    try:
        while True:
            x_feat = []
            y_targets = []
            idx = []
            for i in range(batch_size):
                x_feat.append(row2np(next(features_it)))
                if targets_file is not None:
                    y_targets.append(row2np(next(targets_it)))
                idx.append(gidx)
                gidx += 1
            idx = np.array(idx)
            x_features = x_transformer.transform(np.stack(x_feat))
            if targets_file is None:
                {{return_dict(False, "yield")|indent(16)}}
            else:
                y_class = y_transformer.transform(np.stack(y_targets))
                {{return_dict(True, "yield"|indent(16))}}
    except StopIteration:
        if len(x_feat) == 0:
            raise StopIteration
        idx = np.array(idx)
        if targets_file is None:
            {{return_dict(False, "yield")|indent(12)}}
        else:
            y_class = y_transformer.transform(np.stack(y_targets))
            {{return_dict(True, "yield")|indent(12)}}
{% endif %}



"""Main CLI commands
"""
from __future__ import absolute_import
from __future__ import print_function

import argparse
import json
import sys
import os
from ..utils import parse_json_file_str, cd, makedir_exist_ok, compare_numpy_dict
import kipoi  # for .config module
from kipoi.cli.parser_utils import add_model, add_source, add_dataloader, add_dataloader_main, file_exists, dir_exists
from kipoi.sources import list_subcomponents
from ..data import numpy_collate_concat
# import h5py
# import six
import numpy as np
import pandas as pd
from tqdm import tqdm
from collections import OrderedDict
import logging
from kipoi import writers
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def prepare_batch(dl_batch, pred_batch,
                  keep_inputs=False):
    dl_batch["preds"] = pred_batch

    if not keep_inputs:
        dl_batch.pop("inputs", None)
        dl_batch.pop("targets", None)
    return dl_batch


def cli_test(command, raw_args):
    """Runs test on the model
    """
    assert command == "test"
    # setup the arg-parsing
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description='script to test model zoo submissions. Example usage:\n'
                                     '`kipoi test model/directory`, where `model/directory` is the '
                                     'path to a directory containing a model.yaml file.')
    add_model(parser, source="dir")
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in prediction')
    parser.add_argument("-o", "--output", default=None, required=False,
                        help="Output hdf5 file")
    parser.add_argument("-s", "--skip-expect", action='store_true',
                        help="Skip validating the expected predictions if test.expect field is specified under model.yaml")
    parser.add_argument("-e", "--expect", default=None,
                        help="File path to the hdf5 file of predictions produced by kipoi test -o file.h5 "
                        "or kipoi predict -o file.h5 --keep_inputs. Overrides test.expect in model.yaml")
    args = parser.parse_args(raw_args)
    # --------------------------------------------
    mh = kipoi.get_model(args.model, args.source)

    if not mh._sufficient_deps(mh.dependencies):
        # model requirements should be installed
        logger.warn("Required package '{0}' for model type: {1} is not listed in the dependencies".
                    format(mh.MODEL_PACKAGE, mh.type))

    # Load the test files from model source
    mh.pipeline.predict_example(batch_size=args.batch_size, output_file=args.output)

    if (mh.test.expect is not None or args.expect is not None) \
            and not args.skip_expect and args.output is None:
        if args.expect is not None:
            # `expect` specified from the CLI
            expect = args.expect
        else:
            # `expect` taken from model.yaml
            if isinstance(mh.test.expect, kipoi.specs.RemoteFile):
                # download the file
                output_dir = kipoi.get_source(args.source).get_model_download_dir(args.model)
                makedir_exist_ok(output_dir)
                mh.test.expect = mh.test.expect.get_file(os.path.join(output_dir, 'test.expect.h5'))
            expect = mh.test.expect
        logger.info('Testing if the predictions match the expected ones in the file: {}'.format(expect))
        logger.info('Desired precision (number of matching decimal places): {}'.format(mh.test.precision_decimal))

        # iteratively load the expected file
        expected = kipoi.readers.HDF5Reader(expect)
        expected.open()
        it = expected.batch_iter(batch_size=args.batch_size)
        for i, batch in enumerate(tqdm(it, total=len(expected) // args.batch_size)):
            if i == 0 and ('inputs' not in batch or 'preds' not in batch):
                raise ValueError("test.expect file requires 'inputs' and 'preds' "
                                 "to be specified. Available keys: {}".format(list(expected)))
            pred_batch = mh.predict_on_batch(batch['inputs'])
            # compare to the predictions
            # import ipdb
            # ipdb.set_trace()
            try:
                compare_numpy_dict(pred_batch, batch['preds'], exact=False, decimal=mh.test.precision_decimal)
            except Exception as e:
                logger.error("Model predictions don't match the expected predictions."
                             "expected: {}\nobserved: {}. Exception: {}".format(batch['preds'], pred_batch, e))
                expected.close()
                sys.exit(1)
        expected.close()
        logger.info('All predictions match')
    logger.info('Successfully ran test_predict')


def cli_get_example(command, raw_args):
    """Downloads the example files to the desired directory
    """
    assert command == "get-example"
    # setup the arg-parsing
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description='Get example files')
    add_model(parser, source="kipoi")
    parser.add_argument("-o", "--output", default="example", required=False,
                        help="Output directory where to store the examples. Default: 'example'")
    args = parser.parse_args(raw_args)
    # --------------------------------------------
    mh = kipoi.get_model(args.model, args.source)

    kwargs = mh.default_dataloader.download_example(output_dir=args.output, dry_run=False)

    logger.info("Example files downloaded to: {}".format(args.output))
    logger.info("use the following dataloader kwargs:")
    print(json.dumps(kwargs))


def cli_preproc(command, raw_args):
    """Preprocess:
    - Run the dataloader and store the results to a (hdf5) file
    """
    assert command == "preproc"
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description='Run the dataloader and save the output to an hdf5 file.')
    add_dataloader_main(parser, with_args=True)
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in data loading')
    parser.add_argument("-n", "--num_workers", type=int, default=0,
                        help="Number of parallel workers for loading the dataset")
    parser.add_argument("-o", "--output", required=True,
                        help="Output hdf5 file")
    args = parser.parse_args(raw_args)

    dataloader_kwargs = parse_json_file_str(args.dataloader_args)

    dir_exists(os.path.dirname(args.output), logger)
    # --------------------------------------------
    Dataloader = kipoi.get_dataloader_factory(args.dataloader, args.source)

    dataloader_kwargs = kipoi.pipeline.validate_kwargs(Dataloader, dataloader_kwargs)
    dataloader = Dataloader(**dataloader_kwargs)

    it = dataloader.batch_iter(batch_size=args.batch_size, num_workers=args.num_workers)

    logger.info("Writing to the hdf5 file: {0}".format(args.output))
    writer = writers.HDF5BatchWriter(file_path=args.output)

    for i, batch in enumerate(tqdm(it)):
        # check that the first batch was indeed correct
        if i == 0 and not Dataloader.get_output_schema().compatible_with_batch(batch):
            logger.warn("First batch of data is not compatible with the dataloader schema.")
        writer.batch_write(batch)

    writer.close()
    logger.info("Done!")


def cli_predict(command, raw_args):
    """CLI interface to predict
    """
    assert command == "predict"
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description='Run the model prediction.')
    add_model(parser)
    add_dataloader(parser, with_args=True)
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in prediction')
    parser.add_argument("-n", "--num_workers", type=int, default=0,
                        help="Number of parallel workers for loading the dataset")
    parser.add_argument("-k", "--keep_inputs", action='store_true',
                        help="Keep the inputs in the output file. ")
    parser.add_argument("-l", "--layer",
                        help="Which output layer to use to make the predictions. If specified," +
                        "`model.predict_activation_on_batch` will be invoked instead of `model.predict_on_batch`")
    parser.add_argument('-o', '--output', required=True, nargs="+",
                        help="Output files. File format is inferred from the file path ending. Available file formats are: " +
                        ", ".join(["." + k for k in writers.FILE_SUFFIX_MAP]))
    args = parser.parse_args(raw_args)

    dataloader_kwargs = parse_json_file_str(args.dataloader_args)

    # setup the files
    if not isinstance(args.output, list):
        args.output = [args.output]
    for o in args.output:
        ending = o.split('.')[-1]
        if ending not in writers.FILE_SUFFIX_MAP:
            logger.error("File ending: {0} for file {1} not from {2}".
                         format(ending, o, writers.FILE_SUFFIX_MAP))
            sys.exit(1)
        dir_exists(os.path.dirname(o), logger)
    # --------------------------------------------
    # load model & dataloader
    model = kipoi.get_model(args.model, args.source)

    if args.dataloader is not None:
        Dl = kipoi.get_dataloader_factory(args.dataloader, args.dataloader_source)
    else:
        Dl = model.default_dataloader

    dataloader_kwargs = kipoi.pipeline.validate_kwargs(Dl, dataloader_kwargs)
    dl = Dl(**dataloader_kwargs)

    # setup batching
    it = dl.batch_iter(batch_size=args.batch_size,
                       num_workers=args.num_workers)

    # Setup the writers
    use_writers = []
    for output in args.output:
        writer = writers.get_writer(output, metadata_schema=dl.get_output_schema().metadata)
        if writer is None:
            logger.error("Unknown file format: {0}".format(ending))
            sys.exit()
        else:
            use_writers.append(writer)
    output_writers = writers.MultipleBatchWriter(use_writers)

    # Loop through the data, make predictions, save the output
    for i, batch in enumerate(tqdm(it)):
        # validate the data schema in the first iteration
        if i == 0 and not Dl.get_output_schema().compatible_with_batch(batch):
            logger.warn("First batch of data is not compatible with the dataloader schema.")

        # make the prediction
        if args.layer is None:
            pred_batch = model.predict_on_batch(batch['inputs'])
        else:
            pred_batch = model.predict_activation_on_batch(batch['inputs'], layer=args.layer)

        # write out the predictions, metadata (, inputs, targets)
        output_batch = prepare_batch(batch, pred_batch, keep_inputs=args.keep_inputs)
        output_writers.batch_write(output_batch)

    output_writers.close()
    logger.info('Done! Predictions stored in {0}'.format(",".join(args.output)))


def cli_pull(command, raw_args):
    """Pull the repository
    """
    assert command == "pull"
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description="Downloads the directory" +
                                     " associated with the model.")
    parser.add_argument('model', help='Model name. '
                        '<model> can also refer to a model-group - e.g. if you '
                        'specify MaxEntScan then the dependencies\n'
                        'for MaxEntScan/5prime and MaxEntScan/3prime will be installed')
    add_source(parser)
    args = parser.parse_args(raw_args)

    src = kipoi.config.get_source(args.source)
    sub_models = list_subcomponents(args.model, args.source, which='model')
    if len(sub_models) == 0:
        logger.error("Model {0} not found in source {1}".format(args.model, args.source))
        sys.exit(1)
    if len(sub_models) > 1:
        logger.info("Found {0} models under the model name: {1}. Pulling all of them".
                    format(len(sub_models), args.model))
    for sub_model in sub_models:
        src.pull_model(sub_model)


def cli_init(command, raw_args, **kwargs):
    """Initialize the repository using cookiecutter
    """
    assert command == "init"
    logger.info("Initializing a new Kipoi model")

    print("\nPlease answer the questions below. Defaults are shown in square brackets.\n")
    print("You might find the following links useful: ")
    print("- getting started: http://www.kipoi.org/docs/contributing/01_Getting_started/")
    print("- model_type: http://www.kipoi.org/docs/contributing/02_Writing_model.yaml/#type")
    print("- dataloader_type: http://www.kipoi.org/docs/contributing/04_Writing_dataloader.py/#dataloader-types")
    print("--------------------------------------------\n")

    from cookiecutter.main import cookiecutter
    from cookiecutter.exceptions import FailedHookException

    # Get the information about the current directory
    import inspect
    filename = inspect.getframeinfo(inspect.currentframe()).filename
    this_dir = os.path.dirname(os.path.abspath(filename))
    template_path = os.path.join(this_dir, "../model_template/")

    # remove the pyc files in the template directory
    # bug in cookiecutter: https://github.com/audreyr/cookiecutter/pull/1037
    pyc_file = os.path.join(template_path, "hooks", "pre_gen_project.pyc")
    if os.path.exists(pyc_file):
        os.remove(pyc_file)

    # Create project from the cookiecutter-pypackage/ template
    try:
        out_dir = cookiecutter(template_path, **kwargs)
    except FailedHookException:
        # pre_gen_project.py detected an error in the configuration
        logger.error("Failed to initialize the model")
        sys.exit(1)
    print("--------------------------------------------")
    logger.info("Done!\nCreated the following folder into the current working directory: {0}".format(os.path.basename(out_dir)))


def cli_info(command, raw_args):
    """CLI interface to predict
    """
    assert command == "info"
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description="Prints dataloader" +
                                                 " keyword arguments.")
    add_model(parser)
    add_dataloader(parser, with_args=False)
    args = parser.parse_args(raw_args)

    # --------------------------------------------
    # load model & dataloader
    model = kipoi.get_model(args.model, args.source)

    if args.dataloader is not None:
        dl_info = "dataloader '{0}' from source '{1}'".format(str(args.dataloader), str(args.dataloader_source))
        Dl = kipoi.get_dataloader_factory(args.dataloader, args.dataloader_source)
    else:
        dl_info = "default dataloader for model '{0}' from source '{1}'".format(str(model.name), str(args.source))
        Dl = model.default_dataloader

    print("-" * 80)
    print("Displaying keyword arguments for {0}".format(dl_info))
    print(Dl.print_args())
    print("-" * 80)


def cli_list_plugins(command, raw_args):
    """CLI interface to predict
    """
    assert command == "list_plugins"
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description="Lists available pluging")
    parser.parse_args(raw_args)
    print(kipoi.list_plugins().to_string(index=False, justify="unset"))


def cli_ls(command, raw_args):
    """List all kipoi models
    """
    assert command == "ls"
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description="Lists available models")
    parser.add_argument("group_filter", nargs='?', default='',
                        help="A relative path to the model group used to subset the model list. Use 'all' to show all models")
    parser.add_argument("--tsv", action='store_true',
                        help="Print the output in the tsv format.")
    add_source(parser)
    args = parser.parse_args(raw_args)
    grp = kipoi.get_source(args.source)
    df = grp.list_models()
    ls_helper(df, args.group_filter, args.tsv)


#  split it up for easier testing

def ls_helper(df, group_filter='', tsv=False):
    if group_filter == 'all':
        for m in list(df.model):
            print(m)
    else:
        from kipoi.sources import list_models_by_group
        dfg = list_models_by_group(df, group_filter)
        if dfg is None:
            # print the model list
            models = df.model[df.model.str.contains(group_filter)]
            for m in models:
                print(m)
        else:
            if tsv:
                dfg[['group', 'N_models', 'N_subgroups']].to_csv(sys.stdout, sep='\t', index=False)
            else:
                for i, row in dfg.iterrows():
                    if row.N_subgroups == 0 and row.N_models == 1:
                        print("{}".format(row.group))
                    else:
                        print("{}/ ({})".format(row.group, row.N_models))



from collections import OrderedDict
from related.functions import to_model
from related.types import TypedSequence
from six import string_types


def to_sequence_field_w_str(cls):
    """
    Returns a callable instance that will convert a value to a Sequence.

    :param cls: Valid class type of the items in the Sequence.
    :return: instance of the SequenceConverter.
    """
    class SequenceConverter(object):

        def __init__(self, cls):
            self.cls = cls

        def __call__(self, values):
            values = values or []
            # if values is a string, consider it to be a single argument
            if isinstance(values, str):
                args = [values]
            elif isinstance(values, list):
                args = [to_model(self.cls, value) for value in values]
            else:
                raise ValueError("Field has to be a string or a list of strings")
            return TypedSequence(cls=self.cls, args=args)

    return SequenceConverter(cls)


# has description field

# keyword = "description"
def to_leaf_mapping_field(cls, keyword, key):
    """
    Returns a callable instance that will convert a value to a Sequence.

    :param cls: Valid class type of the items in the Sequence.
    :param key: Attribute name of the key value in each item of cls instance.
    :return: instance of the SequenceConverter.
    """
    # Input = dict
    # if type
    class LeafConverter(object):

        def __init__(self, cls, key):
            self.cls = cls
            self.key = key

        def __call__(self, value, cur_key=None):
            """Cur_key: current key
            """
            if keyword in value:
                # Stop the recursion - the leaf keyword was found
                if cur_key is not None:
                    if self.key in value:
                        assert value[self.key] == cur_key
                    else:
                        value[self.key] = cur_key

                return to_model(self.cls, value)
            else:
                if isinstance(value, list):
                    value = value or []
                    return [self.__call__(v) for v in value]
                elif isinstance(value, dict):
                    kwargs = OrderedDict()
                    for key_value, item in value.items():
                        kwargs[key_value] = self.__call__(item, cur_key=key_value)
                    return kwargs
                else:
                    raise ValueError("Unable to parse: {0}".format(value))

    return LeafConverter(cls, key)


def to_eval_str(value):
    """
    Returns an eval(str(value)) if the value is not None.

    :param value: None or a value that can be converted to a str.
    :return: None or str(value)
    """
    if isinstance(value, string_types):
        value = eval(str(value))

    return value


def identity(x):
    """Simple identity
    """
    return x



from __future__ import absolute_import
from __future__ import print_function

from collections import OrderedDict, Mapping
import sys
import os
import yaml
import kipoi  # for .config module
from .utils import (load_module, cd, merge_dicts, read_pickle, override_default_kwargs,
                    load_obj, inherits_from, infer_parent_class, makedir_exist_ok)
import abc
import six
import numpy as np
import json
import yaml

from .specs import ModelDescription, RemoteFile, DataLoaderImport, download_default_args
from .pipeline import Pipeline
import logging
from distutils.version import LooseVersion

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class BaseModel(object):
    __metaclass__ = abc.ABCMeta

    MODEL_PACKAGE = None

    @abc.abstractmethod
    def predict_on_batch(self, x):
        raise NotImplementedError

    @classmethod
    def _sufficient_deps(cls, deps):
        """Tests it the provided dependencies contain MODEL_PACKAGE

        Args:
          deps: instance of kipoi.specs.Dependencies

        Returns:
          True if cls.MODEL_PACKAGE is listed in the depenencies and False otherwise
        """
        if cls.MODEL_PACKAGE is None:
            return True
        else:
            for d in deps.conda:
                if cls.MODEL_PACKAGE in d:
                    return True
            for d in deps.pip:
                if cls.MODEL_PACKAGE in d:
                    return True
            return False


def get_model(model, source="kipoi", with_dataloader=True):
    """Load the `model` from `source`, as well as the default dataloder to model.default_dataloder.

    # Arguments
      model (str): model name
      source (str):  source name
      with_dataloader (bool): if True, the default dataloader is
        loaded to `model.default_dataloader` and the pipeline at `model.pipeline` enabled.

    # Returns
    - Instance of class inheriting from `kipoi.models.BaseModel` (like `kipoi.models.KerasModel`)
       decorated with additional attributes.

    # Methods
    - **predict_on_batch(x)**: Make model predictions given a batch of data `x`

    # Appended attributes
    - **type** (`str`): model type (class name)
    - **args** (`dict`): model args used to instantiate the model class
    - **info** (`kipoi.specs.Info`): information about the author (etc)
    - **schema** (`kipoi.specs.ModelSchema`): information about the input/outputdata modalities
    - **dependencies** (`kipoi.specs.Dependencies`): class specifying the dependencies.
          (implements `install` method for running the installation)
    - **default_dataloader** (class inheriting from `kipoi.data.BaseDataLoader`): default
           dataloader. None if `with_dataloader=False` was used.
    - **name** (`str`): model name
    - **source** (`str`): model source
    - **source_dir** (`str`): local path to model source storage
    - **postprocessing** (`dict`): dictionary of loaded plugin specifications
    - **pipeline** (`kipoi.pipeline.Pipeline`): handle to a `Pipeline` object

    """
    # TODO - model can be a yaml file or a directory
    if isinstance(source, str):
        source_name = source
        source = kipoi.config.get_source(source)
    else:
        source_name = 'obj'
        source = source

    # pull the model
    source.pull_model(model)
    # get the model directory
    source_dir = source.get_model_dir(model)
    # get model description
    md = source.get_model_descr(model)

    # TODO - is there a way to prevent code duplication here?
    # TODO - possible to inherit from both classes and call the corresponding inits?
    # --------------------------------------------
    # TODO - load it into memory?

    # TODO - validate md.default_dataloader <-> model

    # Load the dataloader
    if with_dataloader:
        # load from python
        if isinstance(md.default_dataloader, DataLoaderImport):
            with cd(source_dir):
                default_dataloader = md.default_dataloader.get()
            default_dataloader.source_dir = source_dir
            # download util links if specified under default & override the default parameters
            override = download_default_args(default_dataloader.args, source.get_dataloader_download_dir(model))
            if override:
                # override default arguments specified under default
                default_dataloader = override_default_kwargs(default_dataloader, override)
        else:
            # load from directory
            # attach the default dataloader already to the model
            if ":" in md.default_dataloader:
                dl_source, dl_path = md.default_dataloader.split(":")
            else:
                dl_source = source
                dl_path = md.default_dataloader

            # allow to use relative and absolute paths for referring to the dataloader
            default_dataloader_path = os.path.join("/" + model, dl_path)[1:]
            default_dataloader = kipoi.get_dataloader_factory(default_dataloader_path,
                                                              dl_source)
    else:
        default_dataloader = None

    model_download_dir = source.get_model_download_dir(model)
    # Read the Model - append methods, attributes to self
    with cd(source_dir):  # move to the model directory temporarily

        # explicitly handle downloading files for TensorFlowModel
        if md.type == 'tensorflow' or md.defined_as == 'kipoi.model.TensorFlowModel':
            output_dir = os.path.join(model_download_dir, "ckp")
            md.args['checkpoint_path'] = _parse_tensorflow_checkpoint_path(md.args['checkpoint_path'], output_dir)

        # download url links if specified under args
        for k in md.args:
            if isinstance(md.args[k], RemoteFile):
                output_dir = os.path.join(model_download_dir, k)
                logger.info("Downloading model arguments {} from {}".format(k, md.args[k].url))
                makedir_exist_ok(output_dir)

                if md.args[k].md5:
                    fname = md.args[k].md5
                else:
                    fname = "file"

                # download the parameters and override the model
                path = md.args[k].get_file(os.path.join(output_dir, fname))
                md.args[k] = path
        if md.type is not None:
            # old API
            if md.type == 'custom':
                Mod = load_model_custom(**md.args)
                assert issubclass(Mod, BaseModel)  # it should inherit from Model
                mod = Mod()
            elif md.type == "pytorch":
                mod = infer_pyt_class(md.args)(**md.args)
            elif md.type in AVAILABLE_MODELS:
                # TODO - this doesn't seem to work
                mod = AVAILABLE_MODELS[md.type](**md.args)
            else:
                raise ValueError("Unsupported model type: {0}. " +
                                 "Model type needs to be one of: {1}".
                                 format(md.type,
                                        ['custom'] + list(AVAILABLE_MODELS.keys())))
        else:
            # new API
            try:
                Mod = load_obj(md.defined_as)
            except ImportError:
                if md.defined_as.startswith("kipoi.model."):
                    # user tried importing some of the available models
                    logger.error("{} is not a valid kipoi model. Available models are: {}\n".format(
                        md.defined_as,
                        ", ".join(["kipoi.model." + str(AVAILABLE_MODELS[k].__name__) for k in AVAILABLE_MODELS])
                    ))
                raise ImportError("Unable to import {}".format(md.defined_as))
            if not inherits_from(Mod, BaseModel):
                raise ValueError("Model {} needs to inherit from kipoi.model.BaseModel".format(md.defined_as))
            mod = Mod(**md.args)
            for k, v in six.iteritems(AVAILABLE_MODELS):
                if isinstance(mod, v):
                    md.type = k
            if md.type is None:
                md.type = 'custom'

    # populate the returned class
    mod.type = md.type
    mod.args = md.args
    mod.info = md.info
    mod.schema = md.schema
    mod.dependencies = md.dependencies
    mod.test = md.test
    mod.default_dataloader = default_dataloader
    mod.name = model
    mod.source = source
    mod.source_name = source_name
    mod.source_dir = source_dir
    # parse the postprocessing module
    mod.postprocessing = md.postprocessing
    if with_dataloader:
        mod.pipeline = Pipeline(model=mod, dataloader_cls=default_dataloader)
    else:
        mod.pipeline = None
    return mod


# ------ individual implementations ----
# each requires a special module to be installed (?)
# - TODO - where to specify those requirements?
#      model: model's relative path / name in the source.
#      2nd column in the `kipoi.list_models()` `pd.DataFrame`.


def load_model_custom(file, object):
    """Loads the custom Model

    # model.yml entry

        ```
        Model:
          type: custom
          args:
            file: model.py
            object: Model
        ```
    """
    return getattr(load_module(file), object)


class GradientMixin(object):
    __metaclass__ = abc.ABCMeta
    allowed_functions = ["sum", "max", "min", "absmax"]

    @abc.abstractmethod
    def input_grad(self, x, filter_idx=None, avg_func=None, layer=None, final_layer=True,
                   selected_fwd_node=None, pre_nonlinearity=False):
        """
        Calculate the input-layer gradient for filter `filter_idx` in layer `layer` with respect to `x`. If avg_func
        is defined average over filters with the averaging function `avg_func`. If `filter_idx` and `avg_func` are both
        not None then `filter_idx` is first applied and then `avg_func` across the selected filters.

        # Arguments
            x: model input
            filter_idx: filter index of `layer` for which the gradient should be returned
            avg_func: String name of averaging function to be applied across filters in layer `layer`
            layer: layer from which backwards the gradient should be calculated
            final_layer: Use the final (classification) layer as `layer`
            selected_fwd_node: None - not supported by KerasModel at the moment
            pre_nonlinearity: Try to use the layer output prior to activation (will not always be possible in an
            automatic way)
        """
        raise NotImplementedError


class LayerActivationMixin():
    @abc.abstractmethod
    def predict_activation_on_batch(self, x, layer, pre_nonlinearity=False):
        """
        Get predictions based on layer output.

        Arguments:
            x: model inputs from dataloader batch
            layer: layer identifier / name. can be integer or string.
            pre_nonlinearity: Assure that output is returned from before the non-linearity. This feature does
            not have to be implemented always (not possible). If not implemented and set to True either raise
            Error or at least warn!
        """
        raise NotImplementedError


class KerasModel(BaseModel, GradientMixin, LayerActivationMixin):
    """Loads the serialized Keras model

    # Arguments
        weights: File path to the hdf5 weights or the hdf5 Keras model
        arch: Architecture json model. If None, `weights` is
    assumed to speficy the whole model
        custom_objects: Python file defining the custom Keras objects
    in a `OBJECTS` dictionary
        backend: Keras backend to use ('tensorflow', 'theano', ...)
        image_dim_ordering: 'tf' or 'th': Whether to use 'tf' ('channels_last')
            or 'th' ('cannels_first') dimension ordering.

    # `model.yml` entry

        ```
        Model:
          type: Keras
          args:
            weights: model.h5
            arch: model.json
            custom_objects: custom_keras_objects.py
        ```
    """

    MODEL_PACKAGE = "keras"

    def __init__(self, weights, arch=None, custom_objects=None, backend=None, image_dim_ordering=None):
        self.backend = backend
        self.image_dim_ordering = image_dim_ordering
        if self.backend is not None and 'KERAS_BACKEND' not in os.environ:
            logger.info("Using Keras backend: {0}".format(self.backend))
            os.environ['KERAS_BACKEND'] = self.backend
        if self.image_dim_ordering is not None:
            import keras.backend as K
            logger.info("Using image_dim_ordering: {0}".format(self.image_dim_ordering))
            K.set_image_dim_ordering(self.image_dim_ordering)
        import keras
        from keras.models import model_from_json, load_model

        if self.backend is not None:
            if keras.backend.backend() != self.backend:
                logger.warn("Keras backend is {0} instead of {1}".
                            format(keras.backend.backend(), self.backend))

        if custom_objects is not None and os.path.exists(custom_objects):
            self.custom_objects = load_module(custom_objects).OBJECTS
        else:
            self.custom_objects = {}

        self.weights = weights
        self.arch = arch

        self.gradient_functions = {}  # contains dictionaries with string reps of filter functions / slices
        self.activation_functions = {}  # contains the activation functions

        if arch is None:
            # load the whole model
            self.model = load_model(weights, custom_objects=self.custom_objects)
            logger.info('successfully loaded the model from {}'.
                        format(weights))
        else:
            # load arch
            with open(arch, "r") as arch:
                self.model = model_from_json(arch.read(),
                                             custom_objects=self.custom_objects)
            logger.info('successfully loaded model architecture from {}'.
                        format(arch))

            # load weights
            self.model.load_weights(weights)
            logger.info('successfully loaded model weights from {}'.
                        format(weights))

    def predict_on_batch(self, x):
        return self.model.predict_on_batch(x)

    def get_layers_and_outputs(self, layer=None, use_final_layer=False, pre_nonlinearity=False):
        """
        Get layers and outputs either by name / index or from the final layer(s).
        If the final layer should be used it has an activation function that is not Linear, then the input to the
        activation function is returned. This check is not performed when `use_final_layer` is False and `layer`
        is being used.

        Arguments:
            layer: layer index (int) or name (non-int)
            use_final_layer:  instead of using `layer` return the final model layer(s) + outputs
        """
        import keras
        sel_outputs = []
        sel_output_dims = []

        def output_sel(layer, output):
            """
            If pre_nonlinearity is true run get_pre_activation_output
            """
            if pre_nonlinearity:
                output = self.get_pre_activation_output(layer, output)[0]  # always has length 1
            return output

        # If the final layer should be used: (relevant for gradient)
        if use_final_layer:
            # Use outputs from the model output layer(s)
            # If the last layer should be selected automatically then:
            # get all outputs from the model output layers
            if isinstance(self.model, keras.models.Sequential):
                selected_layers = [self.model.layers[-1]]
            else:
                selected_layers = self.model.output_layers
            for l in selected_layers:
                for i in range(self.get_num_inbound_nodes(l)):
                    sel_output_dims.append(len(l.get_output_shape_at(i)))
                    sel_outputs.append(output_sel(l, l.get_output_at(i)))

        # If not the final layer then the get the layer by its name / index
        elif layer is not None:
            if isinstance(layer, int):
                selected_layer = self.model.get_layer(index=layer)
            elif isinstance(layer, six.string_types):
                selected_layer = self.model.get_layer(name=layer)
            selected_layers = [selected_layer]
            # get the outputs from all nodes of the selected layer (selecting output from individual output nodes
            # creates None entries when running K.gradients())
            if self.get_num_inbound_nodes(selected_layer) > 1:
                logger.warn("Layer %s has multiple input nodes. By default outputs from all nodes "
                            "are concatenated" % selected_layer.name)
                for i in range(self.get_num_inbound_nodes(selected_layer)):
                    sel_output_dims.append(len(selected_layer.get_output_shape_at(i)))
                    sel_outputs.append(output_sel(selected_layer, selected_layer.get_output_at(i)))
            else:
                sel_output_dims.append(len(selected_layer.output_shape))
                sel_outputs.append(output_sel(selected_layer, selected_layer.output))
        else:
            raise Exception("Either use_final_layer has to be set or a layer name has to be defined.")

        return selected_layers, sel_outputs, sel_output_dims

    @staticmethod
    def get_pre_activation_output(layer, output):
        import keras
        # if the current layer uses an activation function then grab the input to the activation function rather
        # than the output from the activation function.
        # This can lead to confusion if the activation function translates to backend operations that are not a
        # single operation. (Which would also be a misuse of the activation function itself.)
        # suggested here: https://stackoverflow.com/questions/45492318/keras-retrieve-value-of-node-before-activation-function
        if hasattr(layer, "activation") and not layer.activation == keras.activations.linear:
            new_output_ois = []
            if hasattr(output, "op"):
                # TF
                for inp_here in output.op.inputs:
                    new_output_ois.append(inp_here)
            else:
                # TH
                # As of the latest version of Theano this feature is not supported - the activation layer is too
                # diffuse to be handeled here since Theano does not have objects for the activation.
                raise Exception("`get_pre_activation_output` is not supported for Theano models!")
                import theano
                for inp_here in output.owner.inputs:
                    if not isinstance(inp_here, theano.gof.Constant):
                        new_output_ois.append(inp_here)
            if len(new_output_ois) > 1:
                raise Exception("More than one input to activation function of selected layer. No general rule "
                                "implemented for handing those cases. Consider using a linear activation function + a "
                                "non-linear activation layer instead.")
            return new_output_ois
        else:
            return [output]

    @staticmethod
    def get_num_inbound_nodes(layer):
        if hasattr(layer, "_inbound_nodes"):
            # Keras 2.1.5
            return len(layer._inbound_nodes)
        elif hasattr(layer, "inbound_nodes"):
            # Keras 2.0.4
            return len(layer.inbound_nodes)
        else:
            raise Exception("No way to find out about number of inbound Nodes")

    @staticmethod
    def homogenize_filter_slices(filter_slices):
        if isinstance(filter_slices, int):
            filter_slices = (filter_slices,)
        if isinstance(filter_slices, slice):
            filter_slices = (filter_slices,)
        if isinstance(filter_slices, list):
            filter_slices = tuple(filter_slices)
        if isinstance(filter_slices, tuple):
            # Add a 0th dimension for samples if obviously missing, but no information about the actual dimensions
            # is known!
            if len(filter_slices) == 1:
                filter_slices = tuple([slice(None)] + list(filter_slices))
        return filter_slices

    def _get_gradient_function(self, layer=None, use_final_layer=False, pre_nonlinearity=False, filter_slices=None,
                               filter_func=None, filter_func_kwargs=None):
        """
        Get keras gradient function

        # Arguments:
            layer: Layer name or index with respect to which the input gradients should be returned
            use_final_layer: Alternative to `layer`, if the final layer should be used. In this case `layer` can be None.
            filter_slices: Selection of filters in `layer` that should be taken into consideration
            filter_func: Function to be applied on all filters of the selected layer. If both `filter_slices` and
                `filter_func` are defined, then `filter_slices` will be applied first and then `filter_func`.
            filter_func_kwargs: keyword argument dict passed on to `filter_func`
        """
        import keras
        import copy
        from keras.models import Model
        # Generate the gradient functions according to the layer / filter definition
        gradient_function = None

        layer_label = layer
        # Try to use a previously generated gradient function
        if use_final_layer:
            layer_label = "_KIPOI_FINAL_"

        if layer_label is None:
            raise Exception("Either `layer` must be defined or `use_final_layer` set to True.")

        # Cannot query the layer output shape, so only if the slice is an integer or a list of length 1 it is
        # clear that the batch dimension is missing
        if filter_slices is not None:
            filter_slices = self.homogenize_filter_slices(filter_slices)

        if layer_label not in self.gradient_functions:
            self.gradient_functions[layer_label] = {}
        filter_id = str(filter_slices) + "_PNL_" + str(pre_nonlinearity)
        if filter_func is not None:
            filter_id = str(filter_func) + ":" + str(filter_func_kwargs) + ":" + filter_id
        if filter_id in self.gradient_functions[layer_label]:
            gradient_function = self.gradient_functions[layer_label][filter_id]

        if gradient_function is None:
            # model layer outputs wrt which the gradient shall be calculated
            selected_layers, sel_outputs, sel_output_dims = self.get_layers_and_outputs(layer=layer,
                                                                                        use_final_layer=use_final_layer,
                                                                                        pre_nonlinearity=pre_nonlinearity)

            # copy the model input in case learning flag has to appended when using the gradient function.
            inp = copy.copy(self.model.inputs)

            # Now check if layer outputs have to be concatenated (multiple input nodes in the respective layer)
            has_concat_output = False
            if len(sel_outputs) > 1:
                has_concat_output = True
                # Flatten layers in case dimensions don't match
                all_filters_flat = [keras.layers.Flatten()(x) if dim > 2 else x for x, dim in
                                    zip(sel_outputs, sel_output_dims)]
                # A new model has to be generated in order for the concatenated layer output to have a defined layer output
                if hasattr(keras.layers, "Concatenate"):
                    # Keras 2
                    all_filters_merged = keras.layers.Concatenate(axis=-1)(all_filters_flat)
                    gradient_model = Model(inputs=inp, outputs=all_filters_merged)
                else:
                    # Keras 1
                    all_filters_merged = keras.layers.merge(all_filters_flat, mode='concat')
                    gradient_model = Model(input=inp, output=all_filters_merged)
                # TODO: find a different way to get layer outputs...
                # gradient_model.compile(optimizer=self.model.optimizer, loss=self.model.loss)
                gradient_model.compile(optimizer='rmsprop', loss='binary_crossentropy')
                # output of interest for a given gradient
                output_oi = gradient_model.output
            else:
                output_oi = sel_outputs[0]

            # Which subset of the selected layer outputs should be looked at?
            if filter_slices is not None:
                if has_concat_output:
                    logger.warn("Filter slices have been defined for output selection from layers %s, but "
                                "layer outputs of nodes had to be concatenated. This will potentially lead to undesired "
                                "output - please take this concatenation into consideration when "
                                "defining `filter_slices`." % str([l.name for l in selected_layers]))
                output_oi = output_oi[filter_slices]

            # Should a filter function be applied
            if filter_func is not None:
                if filter_func_kwargs is None:
                    filter_func_kwargs = {}
                output_oi = filter_func(output_oi, **filter_func_kwargs)

            if (filter_slices is None) and (filter_func is None):
                raise Exception("Either filter_slices or filter_func have to be set!")

            # generate the actual gradient function
            from keras import backend as K
            saliency = K.gradients(output_oi, inp)

            if self.model.uses_learning_phase and not isinstance(K.learning_phase(), int):
                inp.append(K.learning_phase())

            gradient_function = K.function(inp, saliency)

            # store the generated gradient function:
            self.gradient_functions[layer_label][filter_id] = gradient_function

        return gradient_function

    def _get_feed_input_names(self):
        import keras
        from keras import backend as K
        feed_input_names = None
        if keras.__version__[0] == '1':
            feed_input_names = self.model.input_names
        elif hasattr(keras.engine.training, "_standardize_input_data"):
            from keras.engine.training import _standardize_input_data
            if not hasattr(self.model, "_feed_input_names"):
                if not self.model.built:
                    self.model.build()
            feed_input_names = self.model._feed_input_names
        return feed_input_names

    def _batch_to_list(self, x):
        import keras
        from keras import backend as K
        feed_input_names = self._get_feed_input_names()
        if keras.__version__[0] == '1':
            from keras.engine.training import standardize_input_data as _standardize_input_data
            if not self.model.built:
                self.model.build()
            iis = None
            if hasattr(self.model, "internal_input_shapes"):
                iis = self.model.internal_input_shapes
            elif hasattr(self.model, "model") and hasattr(self.model.model, "internal_input_shapes"):
                iis = self.model.model.internal_input_shapes
            x_standardized = _standardize_input_data(x, feed_input_names,
                                                     iis)
        elif hasattr(keras.engine.training, "_standardize_input_data"):
            from keras.engine.training import _standardize_input_data
            if not hasattr(self.model, "_feed_input_names"):
                if not self.model.built:
                    self.model.build()
            fis = None
            if hasattr(self.model, "_feed_input_shapes"):
                fis = self.model._feed_input_shapes
            x_standardized = _standardize_input_data(x, feed_input_names, fis)
        else:
            raise Exception("This Keras version is not supported!")
        return x_standardized

    def _match_to_input(self, to_match, input):
        feed_input_names = self._get_feed_input_names()
        if isinstance(input, np.ndarray):
            assert len(to_match) == 1
            outputs = to_match[0]
        elif isinstance(input, list):
            # Already in right format
            outputs = to_match
        elif isinstance(input, dict):
            from collections import OrderedDict
            outputs_dict = OrderedDict()
            for k, v in zip(feed_input_names, to_match):
                outputs_dict[k] = v
            outputs = outputs_dict
        return outputs

    def _input_grad(self, x, layer=None, use_final_layer=False, filter_slices=None,
                    filter_func=None, filter_func_kwargs=None, pre_nonlinearity=False):
        """Adapted from keras.engine.training.predict_on_batch. Returns gradients for a single batch of samples.

        # Arguments
            x: Input samples, as a Numpy array.

        # Returns
            Numpy array(s) of predictions.
        """
        import keras
        from keras import backend as K
        x_standardized = self._batch_to_list(x)
        if self.model.uses_learning_phase and not isinstance(K.learning_phase(), int):
            ins = x_standardized + [0.]
        else:
            ins = x_standardized
        gf = self._get_gradient_function(layer, use_final_layer=use_final_layer, filter_slices=filter_slices,
                                         filter_func=filter_func, filter_func_kwargs=filter_func_kwargs,
                                         pre_nonlinearity=pre_nonlinearity)
        outputs = gf(ins)

        # re-format to how the input was:
        return self._match_to_input(outputs, x)

    def input_grad(self, x, filter_idx=None, avg_func=None, layer=None, final_layer=True,
                   selected_fwd_node=None, pre_nonlinearity=False):
        """
        Calculate the input-layer gradient for filter `filter_idx` in layer `layer` with respect to `x`. If avg_func
        is defined average over filters with the averaging function `avg_func`. If `filter_idx` and `avg_func` are both
        not None then `filter_idx` is first applied and then `avg_func` across the selected filters.

        # Arguments
            x: model input
            filter_idx: filter index of `layer` for which the gradient should be returned
            avg_func: String name of averaging function to be applied across filters in layer `layer`
            layer: layer from which backwards the gradient should be calculated
            final_layer: Use the final (classification) layer as `layer`
            selected_fwd_node: None - not supported by KerasModel at the moment
            pre_nonlinearity: Try to use the layer output prior to activation (will not always be possible in an
            automatic way)
        """
        import keras.backend as K
        _avg_funcs = {"sum": K.sum, "min": K.min, "max": K.max, "absmax": lambda x: K.max(K.abs(x))}
        if avg_func is not None:
            assert avg_func in _avg_funcs
            avg_func = _avg_funcs[avg_func]
        else:
            if K._BACKEND == "theano":
                avg_func = _avg_funcs["sum"]
        if selected_fwd_node is not None:
            raise Exception("'selected_fwd_node' is currently not supported for Keras models!")
        return self._input_grad(x, layer=layer, filter_slices=filter_idx, use_final_layer=final_layer,
                                filter_func=avg_func, pre_nonlinearity=pre_nonlinearity)

    def _generate_activation_output_functions(self, layer, pre_nonlinearity):
        import copy
        layer_id = str(layer) + "_" + str(pre_nonlinearity)
        if layer_id in self.activation_functions:
            return self.activation_functions[layer_id]

        # get the selected layers
        selected_layers, sel_outputs, sel_output_dims = self.get_layers_and_outputs(layer=layer,
                                                                                    use_final_layer=False,
                                                                                    pre_nonlinearity=pre_nonlinearity)

        # copy the model input in case learning flag has to appended when using the activation function.
        inp = copy.copy(self.model.inputs)

        # Can't we have multiple outputs for the function?
        output_oi = sel_outputs  # list of outputs should work: https://keras.io/backend/#backend-functions -> backend.function

        from keras import backend as K

        if self.model.uses_learning_phase and not isinstance(K.learning_phase(), int):
            inp.append(K.learning_phase())

        activation_function = K.function(inp, output_oi)

        # store the generated activation function:
        self.activation_functions[layer_id] = activation_function

        return activation_function

    def predict_activation_on_batch(self, x, layer, pre_nonlinearity=False):
        """Adapted from keras.engine.training.predict_on_batch. Returns gradients for a single batch of samples.

        Arguments
            x: Input samples, as a Numpy array.

        Returns
            Numpy array(s) of predictions.
        """
        import keras
        from keras import backend as K
        if keras.__version__[0] == '1':
            from keras.engine.training import standardize_input_data as _standardize_input_data
            x = _standardize_input_data(x, self.model.input_names,
                                        self.model.internal_input_shapes)
        else:
            from keras.engine.training import _standardize_input_data
            x = _standardize_input_data(x, self.model._feed_input_names,
                                        self.model._feed_input_shapes)
        if self.model.uses_learning_phase and not isinstance(K.learning_phase(), int):
            ins = x + [0.]
        else:
            ins = x
        af = self._generate_activation_output_functions(layer, pre_nonlinearity)
        outputs = af(ins)
        return outputs


class PyTorchFwdHook(object):
    def __init__(self):
        self.forward_values = []

    def run_forward_hook(self, module, input, output):
        self.forward_values.append(output)


pt_type_conversions = {
    "FloatTensor": "float",
    "DoubleTensor": "double",
    "HalfTensor": "half",
    "ByteTensor": "byte",
    "CharTensor": "char",
    "ShortTensor": "short",
    "IntTensor": "int",
    "LongTensor": "long"}
pt_type_conversions = {pre + k: v for pre in ["torch.", "torch.cuda."] for k, v in pt_type_conversions.items()}

def infer_pyt_class(kwargs):
    minimum_kwargs_new = (("weights", "module_obj"), ("weights", "module_class"))
    given_kwargs = set(kwargs.keys())
    if any([set(kwargs_new).issubset(given_kwargs) for kwargs_new in minimum_kwargs_new]):
        return PyTorchModel
    else:
        return OldPyTorchModel

class PyTorchModel(BaseModel, GradientMixin, LayerActivationMixin):
    """Loads a pytorch model.

    """

    MODEL_PACKAGE = "pytorch"

    def __init__(self, weights, module_class=None, module_kwargs=None, module_obj=None, module_file=None,
                 auto_use_cuda=True):
        """
        Instantiate a PyTorchModel. The preferred way of instantiating PyTorch models is by using the `load_state_dict`
        method of the model class that specifies the PyTorch model.

        This is in agreement with:
         https://pytorch.org/tutorials/beginner/saving_loading_models.html#saving-loading-model-for-inference

        Arguments:
          weights: file in which the weights are stored
          module_class: name of the PyTorch module class (model class) defined in the `module_file` file. Also the
           `my_module_file.MyModuleClass` is allowed where `my_module_file.py` resides in the same folder as the
           `model.yaml`.
          module_kwargs: If `module_class` is used then kwargs for the module initialisation can be defined here.
          module_obj: name of the PyTorch module object ("model") defined in the `module_file` file. Also
            the `my_module_file.MyModule` is allowed where `my_module_file.py` resides in the same folder as the
           `model.yaml`.
          module_file: path to the python file defining either `module_obj` or `module_class`
          auto_use_cuda: Automatically try to use CUDA if available
        """
        import torch
        from kipoi.utils import load_obj

        if (module_obj is None) and (module_class is None):
            raise Exception("Either 'module_obj' or 'module_class' have to be defined.")

        obj_name = module_class
        if module_obj is not None:
            obj_name = module_obj

        if module_file is not None:
            obj = getattr(load_module(module_file), obj_name)
        else:
            try:
                obj = load_obj(obj_name)
            except ValueError as e:
                raise ValueError("The module file either has to be defined explicitly in `module_file` or implicitly "
                                 "in the `module_class` or `module_obj` arguments. Loading the PyTorchModel failed "
                                 "with: %s" % e.message)

        self.model = obj
        if module_class is not None:
            kwargs = {}
            if module_kwargs is not None:
                if isinstance(module_kwargs, six.string_types):
                    kwargs = yaml.load(module_kwargs)
                else:
                    kwargs = module_kwargs
            self.model = obj(**kwargs)

        self.model.load_state_dict(torch.load(weights))

        if auto_use_cuda and torch.cuda.is_available():
            self.model = self.model.cuda()
            self.use_cuda = True
        else:
            self.use_cuda = False

        # Assuming that model should be used for predictions only
        self.model.eval()

        # Keep all gradient hooks in a list
        self.grad_hooks = []

    @staticmethod
    def correct_neg_stride(x):
        if any([el < 0 for el in x.strides]):
            # pytorch doesn't support negative strides at the moment, copying the numpy array will create a new array
            # with positive strides.
            return x.copy()
        return x

    def _torch_var(self, input, requires_grad=False):
        from torch.autograd import Variable
        if self.use_cuda:
            input = input.cuda()
        out = Variable(input, requires_grad=requires_grad)
        return out

    def _torch_var_to_numpy(self, input):
        if input.is_cuda:
            input = input.cpu()
        return input.data.numpy()

    def _model_is_cuda(self):
        return next(self.model.parameters()).is_cuda

    def predict_on_batch(self, x):
        """
        Input dictionaries will be translated into **kwargs of the `model.forward(...)` call
        Input lists will be translated into *args of the `model.forward(...)` call
        Input np.ndarray will be used as the only argument in a `model.forward(...)` call
        """

        pred, _ = self.np_run_pred(x)
        return self.pred_to_np(pred)

    def numpy_to_torch(self, x, requires_grad=False):
        import torch
        if isinstance(x, np.ndarray):
            # convert to a pytorch tensor and then to a pytorch variable
            input = self._torch_var(torch.from_numpy(self.correct_neg_stride(x)), requires_grad)

        elif isinstance(x, dict):
            # convert all entries in the dict to pytorch variables
            input = {k: self._torch_var(torch.from_numpy(self.correct_neg_stride(x[k])), requires_grad) for k in x}

        elif isinstance(x, list):
            # convert all entries in the list to pytorch variables
            input = [self._torch_var(torch.from_numpy(self.correct_neg_stride(el)), requires_grad) for el in x]

        else:
            raise Exception("Input not supported!")

        return input

    def np_run_pred(self, x, requires_grad=False):
        """
        Input dictionaries will be translated into **kwargs of the `model.forward(...)` call
        Input lists will be translated into *args of the `model.forward(...)` call
        Input np.ndarray will be used as the only argument in a `model.forward(...)` call
        """
        input = self.numpy_to_torch(x, requires_grad=requires_grad)
        if isinstance(x, np.ndarray):
            # convert to a pytorch tensor and then to a pytorch variable
            pred = self.model(input)

        elif isinstance(x, dict):
            # convert all entries in the dict to pytorch variables
            pred = self.model(**input)

        elif isinstance(x, list):
            # convert all entries in the list to pytorch variables
            pred = self.model(*input)

        else:
            raise Exception("Input not supported!")

        return pred, input

    def pred_to_np(self, pred):
        from torch.autograd import Variable
        # convert results back to numpy arrays
        if isinstance(pred, Variable):
            pred_np = self._torch_var_to_numpy(pred)

        elif isinstance(pred, dict):
            pred_np = {k: self._torch_var_to_numpy(pred[k]) for k in pred}

        elif isinstance(pred, list) or isinstance(pred, tuple):
            pred_np = [self._torch_var_to_numpy(el) for el in pred]

        else:
            raise Exception("Model output format not supported!")

        return pred_np

    def get_layer(self, index):
        """
        Get layer (module) based on index: index for sequentials is e.g.: '1.5.1', for models defined as sublcasses of
        nn.Module it's the class object variable names
        """
        # index for sequentials is e.g.: '1.5.1', for models defined as sublcasses of nn.Module it's the class object
        # variable names
        for idx, m in self.model.named_modules():
            if idx == index:
                return m

    def get_layer_id(self, layer):
        for idx, m in self.model.named_modules():
            if m == layer:
                return idx

    @staticmethod
    def extract_module_id(trace_node_obj):
        import re
        import torch
        sqb_restr = r"\[([A-Za-z0-9_]+)\]"
        if LooseVersion(torch.__version__) < LooseVersion('0.4.0'):
            scopeName = trace_node_obj.scopeName()
        else:
            scopeName = trace_node_obj.node().scopeName()
        idx_name = ".".join([re.search(sqb_restr, grp).group(1) for grp in
                             scopeName.split("/") if re.search(sqb_restr, grp) is not None])
        return idx_name

    @staticmethod
    def _is_nonlinear_activation(layer):
        import torch
        import torch.nn.modules.activation as tact
        import inspect
        activation_modules = []
        activation_module_names = []
        for mod_name, mod in tact.__dict__.items():
            if inspect.isclass(mod) and issubclass(mod, torch.nn.Module):
                if mod != torch.nn.Module:
                    activation_modules.append(mod)
                    activation_module_names.append(mod_name)
        # This will also catch instances of subclasses
        return any([isinstance(layer, mod) for mod in activation_modules])

    def _get_trace(self, x):
        import torch
        # Versions of Pytorch prior to '0.4.0':
        if LooseVersion(torch.__version__) < LooseVersion('0.4.0'):
            trace_fn = torch.jit.trace
        else:
            trace_fn = torch.jit.get_trace_graph

        trace = None

        if isinstance(x, np.ndarray):
            trace, _ = trace_fn(self.model, args=(self.numpy_to_torch(x),))
        elif isinstance(x, dict):
            trace, _ = trace_fn(self.model, kwargs=self.numpy_to_torch(x))
        elif isinstance(x, list):
            trace, _ = trace_fn(self.model, args=tuple(self.numpy_to_torch(x)))

        return trace

    def get_last_layers(self, x):
        """
        Returns the model output layers
        x must be a pytorch Variable compatible with the model input
        """
        trace = self._get_trace(x)
        layer_idxs = [self.extract_module_id(n) for n in trace.graph().outputs()]
        return [self.get_layer(i) for i in layer_idxs]

    def get_downstream_layers(self, x, layer_id):
        # layers that are only created in the forward call (e.g. activation layers?) cannot be referred to properly
        raise Exception("No safe graph taversal is implemented yet!")
        layer_output_unames = []  # unique names of layer outputs (data streams)
        trace = self._get_trace(x)

        # Iterate over all modules in the graph and remember the module outputs (so that they can be checked later)
        for mod in trace.graph().nodes():
            if self.extract_module_id(mod) == layer_id:
                layer_output_unames += [n.uniqueName() for n in mod.outputs()]

        # get the model output stream names to check is it is a leaf output
        model_output_unames = [n.uniqueName() for n in trace.graph().outputs()]
        # get the layer ids that receive data from layer_id
        next_layer_ids = []
        for mod in trace.graph().nodes():
            this_layer_inputs = [n.uniqueName() for n in mod.inputs()]
            if any([iuname in layer_output_unames for iuname in this_layer_inputs]):
                next_layer_ids.append(self.extract_module_id(mod))

        if "" in next_layer_ids:
            raise Exception("The model is not compatible with the current implementation")

        # some values are fed back to the layer itself.
        next_layer_ids = [lid for lid in next_layer_ids if lid != layer_id]

        # layers receiving from the given layer, is the layer (also) an output leaf node
        return next_layer_ids, [self.get_layer(i) for i in next_layer_ids], \
            any([iuname in layer_output_unames for iuname in model_output_unames])

    @classmethod
    def _pt_type_match(self, in_data, like, match_cuda=False):
        from torch.autograd import Variable
        if isinstance(like, Variable):
            in_type = like.data.type()
        else:
            in_type = like.type()
        out_data = getattr(in_data, pt_type_conversions[in_type])()
        if match_cuda and like.is_cuda:
            out_data = out_data.cuda()
        return out_data

    @classmethod
    def get_grad_tens(self, forward_values, filter_slices, filter_func):
        import torch
        import six
        if (filter_slices is None) and (filter_func is None):
            raise Exception("Either filter slices or filter function have to be defined")
        if filter_func is not None:
            if not (isinstance(filter_func, six.string_types) and filter_func in self.allowed_functions):
                raise Exception("filter_func has to be a string within %s" % str(self.allowed_functions))

        # perform given operations in numpy, simpler implementation...
        if filter_slices is not None:
            pt_filt_slice = torch.from_numpy(get_filter_array(filter_slices, tuple(forward_values.size()))).byte()
        else:
            pt_filt_slice = torch.ByteTensor(*forward_values.size())
            pt_filt_slice[:] = 1
        if filter_func is not None:
            if filter_func == "sum":
                # don't do anything and keep the filter mask
                pass
            else:
                float_mask = self._pt_type_match(pt_filt_slice, forward_values)
                subset_dataset = forward_values.cpu().data * float_mask
                pt_filt_slice_new = torch.ByteTensor(*pt_filt_slice.size()).zero_()
                if filter_func == "max":
                    # reset so that the masked values are the minimum
                    subset_dataset = subset_dataset + (1 - float_mask) * subset_dataset.min()
                    # Find the maximum output value among the selected
                    for i in range(subset_dataset.shape[0]):
                        pt_filt_slice_new[i] = subset_dataset[i] == subset_dataset[i].max()
                elif filter_func == "min":
                    # reset so that the masked values are the maximum
                    subset_dataset = subset_dataset + (1 - float_mask) * subset_dataset.max()
                    # Find the minimum value among the selected
                    for i in range(subset_dataset.shape[0]):
                        pt_filt_slice_new[i] = subset_dataset[i] == subset_dataset[i].min()
                elif filter_func == "absmax":
                    # Find the absmax value among the selected
                    for i in range(subset_dataset.shape[0]):
                        pt_filt_slice_new[i] = subset_dataset[i] == subset_dataset[i].abs().max()
                pt_filt_slice = pt_filt_slice_new
        pt_filt_slice = self._pt_type_match(pt_filt_slice, forward_values, match_cuda=True)
        return pt_filt_slice

    def _register_fwd_hook(self, layer):
        """
        Install a forward hook on the given layer index
        Returns a PytorchFwdHook object that contains a
        """
        fwd_hook_obj = PyTorchFwdHook()
        removable_hook_obj = layer.register_forward_hook(fwd_hook_obj.run_forward_hook)
        return fwd_hook_obj, removable_hook_obj

    def _input_grad(self, x, layer=None, filter_slices=None, filter_func=None, selected_fwd_node=None):
        import torch
        # Register hooks for the layers
        fwd_hook_obj, removable_hook_obj = self._register_fwd_hook(layer)

        pred, x_in = self.np_run_pred(x, requires_grad=True)

        self.model.zero_grad()

        if selected_fwd_node is not None:
            if isinstance(selected_fwd_node, int):
                grad_concat = fwd_hook_obj.forward_values[selected_fwd_node]
            else:
                raise Exception("'selected_fwd_node' can either be None or an integer indicating the PyTorch model"
                                "forward-iteration of the sepcified layer / module.")
        else:
            # in Keras at the moment the tensors are flattened and then concatenated for every position.
            flat_fwds = [el.view(el.size(0), -1) for el in fwd_hook_obj.forward_values]
            grad_concat = torch.cat(flat_fwds, 1)

        replacement_grad = self.get_grad_tens(grad_concat, filter_slices=filter_slices, filter_func=filter_func)
        grad_concat.backward(gradient=replacement_grad)
        removable_hook_obj.remove()

        def extract_grad(variable_obj):
            vo = variable_obj
            if vo.grad is not None:
                return vo.grad.cpu().data.numpy()
            else:
                ret_arr = np.empty(vo.size())
                ret_arr[:] = np.nan
                return ret_arr

        if isinstance(x_in, torch.autograd.Variable) or isinstance(x_in, torch.Tensor):
            # make sure it is on the cpu, then extract the gradient data as numpy arrays
            grad_out = extract_grad(x_in)

        elif isinstance(x_in, dict):
            # extract gradient values for all dict entries
            from collections import OrderedDict
            grad_out = OrderedDict()
            for k in x:
                grad_out[k] = extract_grad(x_in[k])

        elif isinstance(x_in, list):
            # extract gradient values for all list entries
            grad_out = [extract_grad(el) for el in x_in]

        else:
            raise Exception("Gradient could not be extracted!")

        return grad_out

    def input_grad(self, x, filter_idx=None, avg_func=None, layer=None, final_layer=True,
                   selected_fwd_node=None, pre_nonlinearity=False):
        """
        Calculate the input-layer gradient for filter `filter_idx` in layer `layer` with respect to `x`. If avg_func
        is defined average over filters with the averaging function `avg_func`. If `filter_idx` and `avg_func` are both
        not None then `filter_idx` is first applied and then `avg_func` across the selected filters.

        # Arguments
            x: model input
            filter_idx: filter index of `layer` for which the gradient should be returned
            avg_func: String name of averaging function to be applied across filters in layer `layer`
            layer: layer from which backwards the gradient should be calculated
            final_layer: Use the final (classification) layer as `layer`
            selected_fwd_node: None or integer. If a layer is re-used models may support that the gradient is
            calculated only with respect to one of the incoming edges / nodes.
            pre_nonlinearity: Try to use the layer output prior to activation (will not always be possible in an
            automatic way)
        """

        if layer is not None:
            selected_layers = [self.get_layer(layer)]
        elif final_layer:
            selected_layers = self.get_last_layers(x)
        else:
            raise Exception("Either `layer` must be defined or `final_layer` set to True.")

        if pre_nonlinearity:
            raise Exception("pre_nonlinearity is not implemented for PyTorch models.")

        if len(selected_layers) > 1:
            # This could be implemented by sequentially looping over layers
            raise Exception("Only one layer may be selected at a time!")

        if selected_layers[0] is None:
            raise ValueError("Unable to get layer {}".format(layer))

        return self._input_grad(x, layer=selected_layers[0], filter_slices=filter_idx, filter_func=avg_func,
                                selected_fwd_node=selected_fwd_node)

    def get_upstream_layers(self, x, layer_id):
        # layers that are only created in the forward call (e.g. activation layers?) cannot be referred to properly
        raise Exception("No safe graph taversal is implemented yet!")
        layer_input_unames = []  # unique names of layer outputs (data streams)
        trace = self._get_trace(x)

        # Iterate over all modules in the graph and remember the module inputs (so that they can be checked later)
        for mod in trace.graph().nodes():
            if self.extract_module_id(mod) == layer_id:
                layer_input_unames += [n.uniqueName() for n in mod.inputs()]

        # get the model input stream names to check is it is a leaf input
        model_input_unames = [n.uniqueName() for n in trace.graph().inputs()]

        # get the layer ids that feed data into layer_id
        prev_layer_ids = []
        for mod in trace.graph().nodes():
            this_layer_outputs = [n.uniqueName() for n in mod.outputs()]
            if any([ouname in layer_input_unames for ouname in this_layer_outputs]):
                prev_layer_ids.append(self.extract_module_id(mod))

        # some values are fed back to the layer itself.
        prev_layer_ids = [lid for lid in prev_layer_ids if lid != layer_id]

        # layers feeding into the given layer, is the layer (also) an output leaf node
        return prev_layer_ids, [self.get_layer(i) for i in prev_layer_ids], \
            any([iuname in layer_input_unames for iuname in model_input_unames])

    def predict_activation_on_batch(self, x, layer, pre_nonlinearity=False):
        """Adapted from keras.engine.training.predict_on_batch. Returns gradients for a single batch of samples.

        Arguments
            x: Input samples, as a Numpy array.
        Returns
            List of list of Numpy array(s) of predictions. First level is the hooks (= layers), second level is the
            calls of the hooks (if the layer or its upstream layer is called multiple times in a graph).
        """
        selected_layers = [self.get_layer(layer)]

        if selected_layers[0] is None:
            raise ValueError("Unable to get layer {}".format(layer))

        if pre_nonlinearity:
            raise Exception("pre_nonlinearity is not implemented for PyTorch models.")
            # Currently inactive code because graph traversal is not yet failsafe for all imaginable model architectures
            # if not self._is_nonlinear_activation(selected_layers[0]):
            #    selected_layer_ids, selected_layers, is_leaf = self.get_upstream_layers(x, selected_layers[0])
            #    if len(selected_layer_ids) ==0:
            #        if is_leaf:
            #            raise Exception("Layer '%s' is a nonlinear activation function and is an input leaf node - no "
            #                            "upstream layer could be found!")
            #        else:
            #            raise Exception("Layer '%s' is a nonlinear activation function and no upstream layer could be "
            #                            "found!")

        # Register hooks for the layers
        fwd_hook_objs = []
        removable_hook_objs = []
        for selected_layer in selected_layers:
            fwd_hook_obj, removable_hook_obj = self._register_fwd_hook(selected_layer)
            fwd_hook_objs.append(fwd_hook_obj)
            removable_hook_objs.append(removable_hook_obj)

        # Run full prediction to also
        self.predict_on_batch(x)
        # Remove hook to avoid future use
        [rho.remove() for rho in removable_hook_objs]
        # convert results back and return values.
        # First loop over the hook (= layers) then over the calls of the hooks (inner loop)
        return [[self.pred_to_np(fv) for fv in fho.forward_values] for fho in fwd_hook_objs]

class OldPyTorchModel(PyTorchModel):
    def __init__(self, file=None, build_fn=None, weights=None, auto_use_cuda=True):
        """
        Load model
        `weights`: Path to the where the weights are stored (may also contain model architecture, see below)
        `gen_fn`: Either callable or path to callable that returns a pytorch model object. If `weights` is not None
        then the model weights will be loaded from that file, otherwise it is assumed that the weights are already set
        after execution of `gen_fn()` or the function defined in `gen_fn`.
        Models can be loaded in 2 ways:
        If the model was saved:
        * `torch.save(model, ...)` then the model will be loaded by calling `torch.load(weights)`
        * `torch.save(model.state_dict(), ...)` then another callable has to be passed to arch which returns the
        `model` object, on then `model.load_state_dict(torch.load(weights))` will then be called.
        Where `weights` is the parameter of this function.
        Partly based on: https://stackoverflow.com/questions/42703500/best-way-to-save-a-trained-model-in-pytorch
        """
        logger.warn("You are using the old initialisation of Kipoi's pytorch models! This feature will soon be "
                    "removed. Please convert your model to comply with the new definition of loading 'PyTorchModel's.")
        import torch
        if build_fn is not None:
            if callable(build_fn):
                gen_fn_callable = build_fn

            elif isinstance(build_fn, six.string_types):
                file_path = file
                obj_name = build_fn
                gen_fn_callable = getattr(load_module(file_path), obj_name)

            else:
                raise Exception("gen_fn has to be callable or a string pointing to the callable.")

            # Load model using generator function
            self.model = gen_fn_callable()

            # Load weights
            if weights is not None:
                self.model.load_state_dict(torch.load(weights))

        elif weights is not None:
            # Architecture is stored with the weights (not recommended)
            self.model = torch.load(weights)

        else:
            raise Exception("At least one of the arguments 'weights' or 'gen_fn' has to be set.")

        if auto_use_cuda and torch.cuda.is_available():
            self.model = self.model.cuda()
            self.use_cuda = True
        else:
            self.use_cuda = False

        # Assuming that model should be used for predictions only
        self.model.eval()

        # Keep all gradient hooks in a list
        self.grad_hooks = []


class SklearnModel(BaseModel):
    """Loads the serialized scikit learn model

    # Arguments
        pkl_file: File path to the dumped sklearn file in the pickle format.

    # model.yml entry

        ```
        Model:
          type: sklearn
          args:
            pkl_file: asd.pkl
            predict_method: Which prediction method to use. Available options:
               'predict', 'predict_proba' or 'predict_log_proba'.
        ```
    """

    MODEL_PACKAGE = "scikit-learn"

    def __init__(self, pkl_file, predict_method="predict"):
        self.pkl_file = pkl_file

        from sklearn.externals import joblib
        self.model = joblib.load(self.pkl_file)
        assert predict_method in ['predict_proba', 'predict', 'predict_log_proba']
        assert hasattr(self.model, predict_method)
        self.predict_method = predict_method

    def predict_on_batch(self, x):
        # assert isinstance(x, dict)
        # assert len(x) == 1
        # x = x.popitem()[1]
        return getattr(self.model, self.predict_method)(x)

# --------------------------------------------
# Tensorflow


def get_op_outputs(graph, node_names):
    """Query op names
    """
    if isinstance(node_names, dict):
        return {k: graph.get_operation_by_name(v).outputs[0]
                for k, v in six.iteritems(node_names)}
    elif isinstance(node_names, list):
        return [graph.get_operation_by_name(v).outputs[0]
                for v in node_names]
    elif isinstance(node_names, str):
        return graph.get_operation_by_name(node_names).outputs[0]
    else:
        raise ValueError("node_names has to be dict, list or str. Found: {0}".
                         format(type(node_names)))


def get_filter_array(filter_slices, input_shape):
    def index_is_none(index_el):
        if isinstance(index_el, slice):
            if all([getattr(index_el, att) is None for att in ["start", "stop", "step"]]):
                return True
        return False

    input_dim = len(input_shape)
    pt_filt_slice = np.zeros(input_shape)
    if isinstance(filter_slices, int):
        if input_dim != 2:
            raise Exception(
                "Integer filter slice can only be used on 1D filter, but dimension is: %d" % (input_dim - 1))
        pt_filt_slice[:, filter_slices] = 1
    else:
        if isinstance(filter_slices, slice):
            filter_slices = (filter_slices,)
        if isinstance(filter_slices, list):
            filter_slices = tuple(filter_slices)
        if isinstance(filter_slices, tuple):
            # Add a 0th dimension for samples if missing.
            if len(filter_slices) == input_dim - 1:
                filter_slices = tuple([slice(None)] + list(filter_slices))
            # If dimension is wrong complain
            elif len(filter_slices) != input_dim:
                raise Exception("Filter slice of length %d cannot be applied in a filter of dimension: %d" % (
                    len(filter_slices), input_dim - 1))
            # If sample dimension is not ":" complain
            if not index_is_none(filter_slices[0]):
                raise Exception(
                    "0th (sample) dimension always has to be None. Filter dimension without sample dimension: %d." % (
                        input_dim - 1))
            # Finally apply filter
            pt_filt_slice.__setitem__(filter_slices, 1)
        else:
            raise Exception(
                "filter_slices has to be None or of type integer, tuple or list. Tuples and lists have to contain compatible objects e.g.: slice()-objects.")
    return pt_filt_slice


class TensorFlowModel(BaseModel, GradientMixin, LayerActivationMixin):
    MODEL_PACKAGE = "tensorflow"

    def __init__(self,
                 input_nodes,
                 target_nodes,
                 checkpoint_path,
                 const_feed_dict_pkl=None
                 ):
        """Tensorflow graph

        Args:
          input_nodes: dict(str), list(str) or str: input node names.
            Keys correspond to the values in the feeded data (in schema)
          target_nodes: Same as input_nodes, but for the output node.
            If dict/list, the model will return a dict/list of np.arrays.
          checkpoint_path: Path to the saved model using:
            `saver = tf.train.Saver(); saver.save(checkpoint_path)`
          const_feed_dict_pkl: Constant feed dict stored as a pickle file.
            Values of this dict will get passed every time to feed_dict.
            Hence, const_feed_dict holds required values by the model not
            provided by the Dataloader.
        """
        import tensorflow as tf

        self.input_nodes = input_nodes
        self.target_nodes = target_nodes
        self.checkpoint_path = checkpoint_path
        self.graph = tf.Graph()  # use a fresh graph for the model
        self.sess = tf.Session(graph=self.graph)

        with self.graph.as_default():
            saver = tf.train.import_meta_graph(self.checkpoint_path + '.meta')
            saver.restore(self.sess, self.checkpoint_path)

        self.input_ops = get_op_outputs(self.graph, input_nodes)
        self.target_ops = get_op_outputs(self.graph, target_nodes)

        self.const_feed_dict_pkl = const_feed_dict_pkl
        if self.const_feed_dict_pkl is not None:
            # Load the feed dictionary from the pickle file
            const_feed_dict = read_pickle(self.const_feed_dict_pkl)
            self.const_feed_dict = {self.graph.get_operation_by_name(k).outputs[0]: v
                                    for k, v in six.iteritems(const_feed_dict)}
        else:
            self.const_feed_dict = {}

    def _build_feed_dict(self, x):
        # build feed_dict
        if isinstance(self.input_nodes, dict):
            # dict
            assert isinstance(x, dict)
            feed_dict = {v: x[k] for k, v in six.iteritems(self.input_ops)}
        elif isinstance(self.input_nodes, list):
            # list
            assert isinstance(x, list)
            feed_dict = {v: x[i] for i, v in enumerate(self.input_ops)}
        elif isinstance(self.input_nodes, str):
            # single array
            feed_dict = {self.input_ops: x}
        else:
            raise ValueError

        return feed_dict

    def predict_on_batch(self, x):
        feed_dict = self._build_feed_dict(x)
        return self.sess.run(self.target_ops,
                             feed_dict=merge_dicts(feed_dict, self.const_feed_dict))

    @classmethod
    def get_grad_tens(self, forward_values, filter_slices, filter_func):
        import six
        if (filter_slices is None) and (filter_func is None):
            raise Exception("Either filter slices or filter function have to be defined")
        if filter_func is not None:
            if not (isinstance(filter_func, six.string_types) and filter_func in self.allowed_functions):
                raise Exception("filter_func has to be a string within %s" % str(self.allowed_functions))
        # perform given operations in numpy, simpler implementation...
        if filter_slices is not None:
            pt_filt_slice = get_filter_array(filter_slices, forward_values.shape)
        else:
            pt_filt_slice = np.zeros_like(forward_values)
            pt_filt_slice[:] = 1
        if filter_func is not None:
            if filter_func == "sum":
                # don't do anything and keep the filter mask
                pass
            else:
                float_mask = pt_filt_slice
                subset_dataset = forward_values * float_mask
                pt_filt_slice_new = np.zeros_like(forward_values)
                if filter_func == "max":
                    # reset so that the masked values are the minimum
                    subset_dataset = subset_dataset + (1 - float_mask) * subset_dataset.min()
                    # Find the maximum output value among the selected
                    for i in range(subset_dataset.shape[0]):
                        pt_filt_slice_new[i, ...] = subset_dataset[i, ...] == subset_dataset[i, ...].max()
                elif filter_func == "min":
                    # reset so that the masked values are the maximum
                    subset_dataset = subset_dataset + (1 - float_mask) * subset_dataset.max()
                    # Find the minimum value among the selected
                    for i in range(subset_dataset.shape[0]):
                        pt_filt_slice_new[i, ...] = subset_dataset[i, ...] == subset_dataset[i, ...].min()
                elif filter_func == "absmax":
                    # Find the absmax value among the selected
                    for i in range(subset_dataset.shape[0]):
                        pt_filt_slice_new[i, ...] = subset_dataset[i, ...] == np.abs(subset_dataset[i, ...]).max()
                pt_filt_slice = pt_filt_slice_new * pt_filt_slice
        return pt_filt_slice

    def input_grad(self, x, filter_idx=None, avg_func=None, layer=None, final_layer=True,
                   selected_fwd_node=None, pre_nonlinearity=False):
        """
        Calculate the input-layer gradient for filter `filter_idx` in layer `layer` with respect to `x`. If avg_func
        is defined average over filters with the averaging function `avg_func`. If `filter_idx` and `avg_func` are both
        not None then `filter_idx` is first applied and then `avg_func` across the selected filters.

        # Arguments
            x: model input
            filter_idx: filter index of `layer` for which the gradient should be returned
            avg_func: String name of averaging function to be applied across filters in layer `layer`
            layer: layer from which backwards the gradient should be calculated
            final_layer: Use the final (classification) layer as `layer`
            selected_fwd_node: None - not supported by TensorFlowModel
            pre_nonlinearity: Try to use the layer output prior to activation (will not always be possible in an
            automatic way)
        """

        if selected_fwd_node is not None:
            raise Exception("TensorFlowModel does not support the use of selected_fwd_node!")

        if pre_nonlinearity:
            raise Exception("TensorFlowModel does not support the use of pre_nonlinearity as desired nodes can be "
                            "selected directly!")

        import tensorflow as tf

        feed_dict = self._build_feed_dict(x)

        # get the correct layer with respect to which the gradients should be calculated
        assert isinstance(layer, six.string_types)
        if not final_layer:
            new_target_ops = get_op_outputs(self.graph, layer)
        else:
            new_target_ops = self.target_ops
            if not isinstance(new_target_ops, tf.Tensor):
                raise Exception("Gradients can't be calculated with respect to mutliple layers in TensorFlowModel. "
                                "Please select a single target operation using 'layer'.")

        fwd_values = self.sess.run(new_target_ops,
                                   feed_dict=merge_dicts(feed_dict, self.const_feed_dict))

        # Make sure that input_ops is a list of tf.Tensor objects.
        if isinstance(self.input_ops, dict):
            input_ops = list(self.input_ops.values())
        elif isinstance(self.input_ops, list):
            input_ops = self.input_ops
        elif isinstance(self.input_ops, tf.Tensor):
            input_ops = [self.input_ops]
        else:
            raise ValueError

        # Run the gradient prediction
        # get_grad_tens avoids rebuilding the model which is necessary for TF models!
        grad_op = tf.gradients(new_target_ops, input_ops, name='gradient_%s' % str(layer),
                               grad_ys=self.get_grad_tens(fwd_values, filter_idx, avg_func))
        grad_pred = self.sess.run(grad_op, feed_dict=feed_dict)

        # Format the output so match the input
        if isinstance(self.input_nodes, dict):
            # dict
            grad_pred_formatted = {}
            for op, ret_val in zip(input_ops, grad_pred):
                for k, v in six.iteritems(self.input_ops):
                    if v == op:
                        grad_pred_formatted[k] = ret_val
        elif isinstance(self.input_nodes, list):
            # list
            grad_pred_formatted = []
            for op, ret_val in zip(input_ops, grad_pred):
                for v in self.input_ops:
                    if v == op:
                        grad_pred_formatted.append(ret_val)
        elif isinstance(self.input_nodes, str):
            # single array
            grad_pred_formatted = grad_pred[0]
        else:
            raise ValueError

        return grad_pred_formatted

    def predict_activation_on_batch(self, x, layer, pre_nonlinearity=False):
        """
        Get predictions based on layer output.

        Arguments:
            x: model inputs from dataloader batch
            layer: layer identifier / name. can be integer or string.
            pre_nonlinearity: Not implemented.
        """
        feed_dict = self._build_feed_dict(x)
        new_target_ops = get_op_outputs(self.graph, layer)
        return self.sess.run(new_target_ops,
                             feed_dict=merge_dicts(feed_dict, self.const_feed_dict))


def _parse_tensorflow_checkpoint_path(ckp_path, output_dir):
    """Parse and download tensorflow's checkpoint_path
    """
    if not isinstance(ckp_path, str):
        # need to handle the special case
        error_message = "checkpoint_path needs to be either a string " + \
                        "checkpoint_path needs to be either or a dictionary with " + \
                        "keys: meta, index, data"
        if not isinstance(ckp_path, Mapping):
            raise ValueError(error_message + "\n detected class {}".format(type(ckp_path)))
        if set(ckp_path.keys()) != {'meta', 'index', 'data'}:
            raise ValueError(error_message + "\n detected keys {}".format(set(ckp_path.keys())))

        # either all are string or all are remote paths
        types = {type(ckp_path[k]) for k in ckp_path}
        if len(types) != 1:
            raise ValueError("All types in checkpoint_path need to be the same. Found: {}".format(types))
        if not (isinstance(ckp_path['meta'], str) or isinstance(ckp_path['meta'], RemoteFile)):
            raise ValueError("Values of the checkpoint_path ckp_path need to be either "
                             "str or RemoteFile. Found: {}".format(ckp_path['meta']))
        if isinstance(ckp_path['meta'], RemoteFile):
            # download files
            makedir_exist_ok(output_dir)
            ckp_path['meta'] = ckp_path['meta'].get_file(os.path.join(output_dir, "model.meta"))
            ckp_path['index'] = ckp_path['index'].get_file(os.path.join(output_dir, "model.index"))
            ckp_path['data'] = ckp_path['data'].get_file(os.path.join(output_dir, "model.data-00000-of-00001"))

        if isinstance(ckp_path['meta'], str):
            assert ckp_path['meta'].endswith(".meta")
            assert ckp_path['index'].endswith(".index")
            assert ckp_path['data'].endswith(".data-00000-of-00001")
            # figure out the prefix
            ckp_path_prefix = ckp_path['meta'][:-5]
            assert ckp_path['data'].startswith(ckp_path_prefix)
            assert ckp_path['index'].startswith(ckp_path_prefix)

            return ckp_path_prefix
        else:
            raise ValueError("Values of the checkpoint_path ckp_path need to be either "
                             "str or RemoteFile. Found: {}".format(ckp_path['meta']))
    else:
        return ckp_path

# --------------------------------------------


AVAILABLE_MODELS = OrderedDict([("keras", KerasModel),
                                ("pytorch", PyTorchModel),
                                ("sklearn", SklearnModel),
                                ("tensorflow", TensorFlowModel)])
# "custom": load_model_custom}



"""Configuration management for Kipoi

Following the Keras configuration management:
https://github.com/fchollet/keras/blob/6f3e6bb6fc97e706f37dc078ae821f490b78035b/keras/backend/__init__.py#L14-L43
"""
from __future__ import absolute_import
from __future__ import print_function

import os
from collections import OrderedDict
import pandas as pd
import six
from .sources import load_source, GitSource, GitLFSSource, GithubPermalinkSource, LocalSource
from .utils import yaml_ordered_dump, yaml_ordered_load, du
import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
# --------------------------------------------


_kipoi_base_dir = os.path.expanduser('~')
if not os.access(_kipoi_base_dir, os.W_OK):
    _kipoi_base_dir = '/tmp'

_kipoi_dir = os.path.join(_kipoi_base_dir, '.kipoi')

# model source container
_MODEL_SOURCES = {}


def model_sources():
    if _MODEL_SOURCES:
        return _MODEL_SOURCES
    else:
        # default model source
        return {
            "kipoi": GitSource(remote_url="https://github.com/kipoi/models.git",
                               local_path=os.path.join(_kipoi_dir, "models/"),
                               auto_update=True),
            "github-permalink": GithubPermalinkSource(local_path=os.path.join(_kipoi_dir, "github-permalink/")),
        }


def model_sources_dict():
    return OrderedDict([(k, v.get_config())
                        for k, v in six.iteritems(model_sources())])


def set_model_sources(_model_sources):
    global _MODEL_SOURCES

    _MODEL_SOURCES = _model_sources


def get_source(source):
    """Get the source object

    # Arguments
        source (str): source string

    # Returns
        Source child class instance: like `kipoi.sources.GitLFSSource`
    """
    if source in model_sources():
        return model_sources()[source]
    else:
        raise ValueError("source={0} needs to be in model_sources()\n".format(source) +
                         "available sources: {0}".
                         format(", ".join(list(model_sources().keys()))))


def add_source(name, obj):
    """Append a custom source to global model_sources

    # Arguments
      name: source name
      obj: source object. Can be a dictionary or a Source instance (say `kipoi.sources.LocalSource("mydir/")`).

    """
    if isinstance(obj, dict):
        # parse the object
        obj = load_source(obj, name)
    c_dict = model_sources()
    c_dict.update({name: obj})
    set_model_sources(c_dict)


def list_sources():
    """Returns a `pandas.DataFrame` of possible sources
    """
    def src2dict(k, s):
        lm = s.list_models()
        return OrderedDict([("source", k),
                            ("type", s.TYPE),
                            ("location", s.local_path),
                            ("local_size", du(s.local_path)),
                            ("n_models", len(lm)),
                            ("n_dataloaders", len(lm)),  # TODO - update
                            # last_updated=TODO - implement?
                            ])
    return pd.DataFrame([src2dict(k, s) for k, s in six.iteritems(model_sources()) if k != "dir"])


def list_models(sources=None):
    """List models

    # Arguments
        sources: list of model sources to use. If None, use all

    # Returns
        pandas.DataFrame
    """
    def get_df(source_name, source):
        df = source.list_models()
        df.insert(0, "source", source_name)
        return df

    if sources is None:
        sources = model_sources()

    pd_list = []
    for name, source in six.iteritems(sources):
        if name != "dir":
            pd_list.append(get_df(name, source))

    return pd.concat(pd_list)[pd_list[0].columns]


def list_dataloaders(sources=None):
    """List dataloaders

    # Arguments
        sources: list of model sources to use. If None, use all

    # Returns
        pandas.DataFrame
    """
    def get_df(source_name, source):
        df = source.list_dataloaders()
        df.insert(0, "source", source_name)
        return df

    if sources is None:
        sources = model_sources()

    pd_list = []
    for name, source in six.iteritems(sources):
        if name != "dir":
            pd_list.append(get_df(name, source))

    return pd.concat(pd_list)[pd_list[0].columns]


# Attempt to read Kipoi config file.
_config_path = os.path.expanduser(os.path.join(_kipoi_dir, 'config.yaml'))
if os.path.exists(_config_path):
    try:
        _config = yaml_ordered_load(open(_config_path))
    except ValueError:
        logger.warn("Unable to parse the config file: {0}. Using default config".format(_config_path))
        _model_sources = model_sources()
    else:
        _model_sources = _config['model_sources']
    # dict  -> Source class
    if "dir" in _model_sources:
        raise ValueError("'dir' is a protected key name in model_sources" +
                         " and hence can't be used")

    _model_sources = OrderedDict([(k, load_source(v, k))
                                  for k, v in six.iteritems(_model_sources)])
    assert isinstance(_model_sources, OrderedDict)
    set_model_sources(_model_sources)


# Save config file, if possible.
if not os.path.exists(_kipoi_dir):
    try:
        os.makedirs(_kipoi_dir)
    except OSError:
        # Except permission denied and potential race conditions
        # in multi-threaded environments.
        pass

# Writing the file
if not os.path.exists(_config_path):
    logger.info("{0} doesn't exist. Creating the config file".format(_config_path))
    _config = {
        'model_sources': model_sources_dict(),
    }
    try:
        with open(_config_path, 'w') as f:
            f.write(yaml_ordered_dump(_config, indent=4, default_flow_style=False))
    except IOError:
        # Except permission denied.
        pass


# Add dir as a valid source
add_source("dir", LocalSource(None))

if "github-permalink" not in model_sources():
    # backcompatibility with old versions
    add_source("github-permalink",
               GithubPermalinkSource(local_path=os.path.join(_kipoi_dir, "github-permalink/")))



from __future__ import absolute_import
from __future__ import print_function

import os
import os.path
import hashlib
import errno
from tqdm import tqdm
import imp
import six
import pickle
import glob
import os
import sys
import subprocess
import numpy as np
import functools
import yaml
from collections import OrderedDict
from contextlib import contextmanager
import inspect
import logging
import collections
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# recursive get and setattr
# https://stackoverflow.com/a/31174427
def rgetattr(obj, attr, *args):
    """Recursively get attributes:
    rgetattr(obj, 'attr.subattr')
    """
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return functools.reduce(_getattr, [obj] + attr.split('.'))


def rsetattr(obj, attr, val):
    """Recursively set attributes:
    rsetattr(obj, 'attr.subattr', 10)
    """
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


# temporarily add an object to path
class add_sys_path(object):
    """
    """

    def __init__(self, path):
        self.path = path

    def __enter__(self):
        sys.path.insert(0, self.path)

    def __exit__(self, exc_type, exc_value, traceback):
        sys.path.remove(self.path)


def load_obj(obj_import):
    """Load object from string
    """
    import importlib
    if "." not in obj_import:
        raise ValueError("Object descripiton needs to be of the form: "
                         "module.submodule.Object. currently lacking a dot (.)")

    with add_sys_path(os.getcwd()):
        module_name, obj_name = obj_import.split(".", 1)
        # manually run the import (don't rely on importlib.import_module)
        # the latter was caching modules which caused trouble when
        # loading multiple modules of the same kind
        fp, pathname, description = imp.find_module(module_name)
        e = None
        try:
            module = imp.load_module(module_name, fp, pathname, description)
            obj = rgetattr(module, obj_name)  # recursively get the module
        except Exception as e:
            if fp:
                fp.close()
            raise ImportError("object {} couldn't be imported. Error {}".format(obj_import, str(e)))
        finally:
            # Since we may exit via an exception, close fp explicitly.
            if fp:
                fp.close()
        # module = importlib.import_module(module_name)
    return obj


def load_module(path, module_name=None):
    """Load python module from file

    Args:
       path: python file path
       module_name: import as `module_name` name. If none, use `path[:-3]`
    """
    assert path.endswith(".py")
    if module_name is None:
        module_name = os.path.basename(path)[:-3]  # omit .py

    logger.debug("loading module: {0} as {1}".format(path, module_name))
    if sys.version_info[0] == 2:
        import imp
        # add the directory to system's path - allows loading submodules
        with add_sys_path(os.path.join(os.path.dirname(module_name))):
            module = imp.load_source(module_name, path)
    elif sys.version_info[0] == 3:
        """
        import importlib.machinery
        loader = importlib.machinery.SourceFileLoader
        handle = loader.load_module
        """
        if sys.version_info[1] == 4:
            # way 1 (=python3.4)
            from importlib.machinery import SourceFileLoader
            module = SourceFileLoader(module_name, path).load_module()
        if sys.version_info[1] >= 5:
            # way 2 (>=python3.5)
            import importlib.util

            # alternative
            import types
            loader = importlib.machinery.SourceFileLoader(module_name, path)
            module = types.ModuleType(loader.name)
            loader.exec_module(module)

            # module_spec_ = importlib.util.spec_from_file_location(module_name, path)
            # module = importlib.util.module_from_spec(module_spec_)
            # module_spec_.loader.exec_module(module)
        else:
            raise RuntimeError(
                'dynamic loading of preprocessor module is not implemented for python3!')
    return module


def inherits_from(cls, parent):
    """Check if an object interits from the parent at some point
    """
    for x in inspect.getmro(cls):
        if x == parent:
            return True
    return False


def infer_parent_class(cls, class_dict):
    """Figure out the parent class
    """
    type_inferred = None
    for dl_type in reversed(class_dict):
        dl_cls = class_dict[dl_type]
        if inherits_from(cls, dl_cls):
            return dl_type
    return type_inferred


def pip_install_requirements(requirements_fname):
    if os.path.exists(requirements_fname):  # install dependencies
        logger.info('Running pip install -r {}...'.format(requirements_fname))
        subprocess.call(['pip', 'install', '-r', requirements_fname])
    else:
        logger.info('requirements.txt not found under {}'.format(requirements_fname))


def compare_numpy_dict(a, b, exact=True, decimal=7):
    """
    Compare two recursive numpy dictionaries or lists
    """
    if type(a) != type(b) and type(a) != np.ndarray and type(b) != np.ndarray:
        return False

    # Compare two dictionaries
    if type(a) == dict and type(b) == dict:
        if not a.keys() == b.keys():
            return False
        for key in a.keys():
            res = compare_numpy_dict(a[key], b[key], exact)
            if not res:
                print("false for key = ", key)
                return False
        return True

    # compare two lists
    if type(a) == list and type(b) == list:
        assert len(a) == len(b)
        return all([compare_numpy_dict(a[i], b[i], exact=exact)
                    for i in range(len(a))])

    # if type(a) == np.ndarray and type(b) == np.ndarray:
    if type(a) == np.ndarray or type(b) == np.ndarray:
        if exact:
            return (a == b).all()
        else:
            return np.testing.assert_almost_equal(a, b, decimal=decimal)

    if a is None and b is None:
        return True

    raise NotImplementedError


def parse_json_file_str(extractor_args):
    """Parse a string either as a json string or
    as a file path to a .json file
    """
    extractor_args = extractor_args.strip("'").strip('"')
    if extractor_args.startswith("{") or extractor_args.endswith("}"):
        logger.debug("Parsing the extractor_args as a json string")
        return yaml.load(extractor_args)
    else:
        if not os.path.exists(extractor_args):
            raise ValueError("File path: {0} doesn't exist".format(extractor_args))
        logger.debug("Parsing the extractor_args as a json file path")
        with open(extractor_args, "r") as f:
            return yaml.load(f.read())


# https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts


def yaml_ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass

    def dict_constructor(loader, node):
        return OrderedDict(loader.construct_pairs(node))

    _mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
    OrderedLoader.add_constructor(_mapping_tag, dict_constructor)
    return yaml.load(stream, OrderedLoader)


def yaml_ordered_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
    class OrderedDumper(Dumper):
        pass

    def dict_representer(dumper, data):
        return dumper.represent_dict(six.iteritems(data))
    OrderedDumper.add_representer(OrderedDict, dict_representer)
    return yaml.dump(data, stream, OrderedDumper, **kwds)


@contextmanager
def cd(newdir):
    """Temporarily change the directory
    """
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def getargs(x):
    """Get function arguments
    """
    if sys.version_info[0] == 2:
        if inspect.isfunction(x):
            return set(inspect.getargspec(x).args)
        else:
            # skip the self parameter
            return set(inspect.getargspec(x.__init__).args[1:])
    else:
        return set(inspect.signature(x).parameters.keys())


def _get_arg_name_values(fn_cls):
    """Get the function/class default argument list (and their values)

    Args:
      fn_cls: function or a class. In the class case,
          arguments for  `__init__` are returned
    """
    if sys.version_info[0] == 2:
        getargspec = inspect.getargspec
    else:
        getargspec = inspect.getfullargspec

    if inspect.isfunction(fn_cls):
        args = getargspec(fn_cls).args
        values = fn_cls.__defaults__
    else:
        # skip the self parameter
        args = getargspec(fn_cls.__init__).args[1:]
        values = fn_cls.__init__.__defaults__
    return args, values


def default_kwargs(fn_cls):
    """Get the default kwargs of a function
    """
    args, values = _get_arg_name_values(fn_cls)
    if values is None:
        return {}
    return {args[-i - 1]: values[-i - 1] for i in range(len(values))}


def copy_func(f, name=None):
    """Return a function with same code, globals, defaults, closure, and name (or provide a new name)
    Implemented by: https://stackoverflow.com/a/30714299/7529152

    # Arguments
      f: function
      name: new function name

    # Returns
      function  copy
    """
    import types
    fn = types.FunctionType(f.__code__, f.__globals__, name or f.__name__,
                            f.__defaults__, f.__closure__)
    # in case f was given attrs (note this dict is a shallow copy):
    fn.__dict__.update(f.__dict__)
    return fn


def override_default_kwargs(fn_cls, kwargs):
    """Override default kwargs in fn_cls. It keeps the original
    function / class intact.

    # Arguments
      fn_cls: function or a class

    # Returns
      new function or a class with the original attributes overriden
    """
    if inspect.isfunction(fn_cls):
        # make a copy of the object
        fn_cls = copy_func(fn_cls)
    else:
        def factory(BaseClass):
            """Copy the class and copy also the __init__
            """
            class NewClass(BaseClass):
                __init__ = copy_func(BaseClass.__init__)
            NewClass.__name__ = "Overridden" + BaseClass.__name__
            if sys.version_info[0] == 3:
                NewClass.__qualname__ = "Overridden" + BaseClass.__qualname__
            return NewClass
        fn_cls = factory(fn_cls)

    args, values = _get_arg_name_values(fn_cls)
    # check that all kwargs are specified
    for k in kwargs:
        if k not in args:
            raise ValueError("argument '{}' not specified in "
                             "function/class.__init__ {} with args: {}".format(k, fn_cls, args))

    # set the appropriate args
    out = []
    for i, k in enumerate(args[-len(values):]):
        if k in kwargs:
            out.append(kwargs[k])
        else:
            out.append(values[i])
    new_values = tuple(out)

    if not inspect.isfunction(fn_cls):
        if sys.version_info[0] == 2:
            fn_cls.__init__.__func__.__defaults__ = new_values
        else:
            fn_cls.__init__.__defaults__ = new_values
    else:
        fn_cls.__defaults__ = new_values
    return fn_cls


def read_yaml(path):
    with open(path) as f:
        return yaml.load(f)


def cmd_exists(cmd):
    """Check if a certain command exists
    """
    return subprocess.call("type " + cmd, shell=True,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0


def lfs_installed(raise_exception=False):
    """Check if git lfs is installed localls
    """
    ce = cmd_exists("git-lfs")
    if not ce:
        if raise_exception:
            raise OSError("git-lfs not installed")
        else:
            logger.warn("git-lfs not installed")
    return ce


def get_file_path(file_dir, basename, extensions=[".yml", ".yaml"],
                  raise_err=True):
    """Get the file path allowing for multiple file extensions
    """
    for ext in extensions:
        path = os.path.join(file_dir, basename + ext)
        if os.path.exists(path):
            return path
    if raise_err:
        raise ValueError("File path doesn't exists: {0}/{1}{2}".
                         format(file_dir, basename, set(extensions)))
    else:
        return None


def du(path):
    """disk usage in human readable format (e.g. '2,1GB')"""
    try:
        return subprocess.check_output(['du', '-sh', path]).split()[0].decode('utf-8')
    except Exception:
        return "NA"


class Slice_conv:

    def __getitem__(self, key):
        return key


def unique_list(seq):
    """Make a list unique and preserve the elements order

    Modified version of Dave Kirby solution
    """
    seen = set()
    return [x for x in seq if x not in seen and not seen.add(x)]


def read_txt(file_path, comment_str="#"):
    """Txt file reader that ignores comments and
    empty lines
    """
    out = []
    with open(file_path) as f:
        for line in f:
            line = line.partition(comment_str)[0]
            line = line.strip()
            if len(line) > 0:
                out.append(line)
    return out


def read_pickle(f):
    with open(f, "rb") as f:
        return pickle.load(f)


def merge_dicts(x, y):
    """https://stackoverflow.com/questions/38987/how-to-merge-two-dictionaries-in-a-single-expression
    """
    z = x.copy()   # start with x's keys and values
    z.update(y)    # modifies z with y's keys and values & returns None
    return z


def list_files_recursively(root_dir, basename, suffix='y?ml'):
    """search for filenames matching the pattern: {root_dir}/**/{basename}.{suffix}
    """
    root_dir = os.path.join(root_dir, "")  # make sure root dir ends with "/"
    # TODO - implement skip
    if sys.version_info >= (3, 5):
        return [filename[len(root_dir):] for filename in
                glob.iglob(root_dir + '**/{0}.{1}'.format(basename, suffix), recursive=True)]
    else:
        # TODO - implement skip
        import fnmatch
        return [os.path.join(root, filename)[len(root_dir):]
                for root, dirnames, filenames in os.walk(root_dir)
                for filename in fnmatch.filter(filenames, '{0}.{1}'.format(basename, suffix))]


def map_nested(dd, fn):
    """Map a function to a nested data structure (containing lists or dictionaries

    Args:
      dd: nested data structure
      fn: function to apply to each leaf
    """
    if isinstance(dd, collections.Mapping):
        return {key: map_nested(dd[key], fn) for key in dd}
    elif isinstance(dd, collections.Sequence):
        return [map_nested(x, fn) for x in dd]
    else:
        return fn(dd)


def take_first_nested(dd):
    """Get a single element from the nested list/dictionary

    Args:
      dd: nested data structure

    Example: take_first_nested({"a": [1,2,3], "b": 4}) == 1
    """
    if isinstance(dd, collections.Mapping):
        return take_first_nested(six.next(six.itervalues(dd)))
    elif isinstance(dd, collections.Sequence):
        return take_first_nested(dd[0])
    else:
        return dd


class classproperty(object):
    """https://stackoverflow.com/questions/128573/using-property-on-classmethods
    Allow using @classproperty
    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


def is_subdir(path, directory):
    """Check if the path is in a particular directory

    Example:

    In [102]: is_subdir("/a/b/c", '/a/b')
    Out[105]: True

    In [106]: is_subdir("/a/b/c", '/a/c')
    Out[106]: False
    """
    path = os.path.realpath(path)
    directory = os.path.realpath(directory)
    relative = os.path.relpath(path, directory)
    return not (relative == os.pardir or relative.startswith(os.pardir + os.sep))


def relative_path(full_path, parent_subpath):
    """Get the relative path

    Args:
      path: long path: example /a/b/c
      parent_subpath: sub-directory.

    Example:

    In [78]: relative_path("/a/b/c", '/a')
    Out[78]: 'b/c'

    In [79]: relative_path("/a/b/c", '/a/')
    Out[79]: 'b/c'
    """
    full_path = os.path.realpath(full_path)
    assert parent_subpath != ""
    parent_subpath = os.path.realpath(parent_subpath)
    relative = os.path.relpath(full_path, parent_subpath)
    return relative


def recursive_dict_parse(d, key, fn):
    """Recursively parse the dictionary and apply
    fn once a child dictionary with a key has been observed

    Args:
      d: nested data structure
      key: dictionary key to watch.
      fn: when a dict with `key` is found, apply a function
         to this dictionary
    """
    if isinstance(d, collections.Mapping):
        if key in d:
            return fn(d)
        else:
            return OrderedDict([(k, recursive_dict_parse(v, key, fn)) for k, v in six.iteritems(d)])
    elif isinstance(d, list):
        return [recursive_dict_parse(v, key, fn) for v in d]
    else:
        # nothing to iterate over. stop recursion
        return d


def makedir_exist_ok(dirpath):
    """Python2 support for os.makedirs(.., exist_ok=True)
    """
    try:
        os.makedirs(dirpath)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise OSError(str(e))



"""Utility functions for CLI parsing
"""
from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import kipoi


# parsers

def add_source(parser, default="kipoi"):
    parser.add_argument('--source', default=default,
                        choices=list(kipoi.config.model_sources().keys()),
                        help='Model source to use (default={}). Specified in ~/.kipoi/config.yaml'.format(default) +
                        " under model_sources. " +
                        "When 'dir' is used, use the local directory path when specifying the model/dataloader.")


def add_model(parser, source="kipoi"):
    parser.add_argument('model', help='Model name.')
    add_source(parser, default=source)


def add_dataloader_main(parser, with_args=True):
    parser.add_argument('dataloader', help='Dataloader name.')
    parser.add_argument('--source', default="kipoi",
                        choices=list(kipoi.config.model_sources().keys()),
                        help='Dataloader source to use. Specified in ~/.kipoi/config.yaml' +
                        " under model_sources. " +
                        "'dir' is an additional source referring to the local folder.")
    if with_args:
        parser.add_argument('--dataloader_args',
                            help="Dataloader arguments either as a json string:'{\"arg1\": 1} or " +
                            "as a file path to a json file")


def add_dataloader(parser, with_args=True):
    parser.add_argument('--dataloader', default=None,
                        help="Dataloader name. If not specified, the model's default" +
                        "DataLoader will be used")
    parser.add_argument('--dataloader_source', default="kipoi",
                        help="Dataloader source")

    if with_args:
        parser.add_argument('--dataloader_args',
                            help="Dataloader arguments either as a json string:" +
                            "'{\"arg1\": 1} or as a file path to a json file")


# Multiple models/dataloaders
def add_env_args(parser, source="kipoi"):
    parser.add_argument('model', nargs="+",
                        help='Model name(s). You can use <source>::<model> to use models from different sources. \n'
                        '<model> can also refer to a model-group - e.g. if you specify MaxEntScan then the dependencies\n'
                        'for MaxEntScan/5prime and MaxEntScan/3prime will be installed')
    add_source(parser, default=source)
    parser.add_argument('--dataloader', default=[], nargs="+",
                        help="Dataloader name(s). If not specified, the model's default " +
                        "Dataloader will be used. You can use <source>::<dataloader> to use dataloaders from different sources\n"
                        "As for the --model tag, you can specify whole dataloader groups.")
    parser.add_argument("--vep", action="store_true",
                        help="Include also the dependencies for variant effect prediction")
    parser.add_argument("--gpu", action="store_true",
                        help="Use gpu-compatible dependencies. Example: instead " +
                        "of using 'tensorflow', 'tensorflow-gpu' will be used")


# other utils
def file_exists(fpath, logger):
    if not os.path.exists(fpath):
        logger.error("File {0} doesn't exist".format(fpath))
        sys.exit(1)


def dir_exists(dirname, logger):
    if not os.path.exists(os.path.abspath(dirname)):
        logger.error("Directory {0} doesn't exist".format(dirname))
        sys.exit(1)


def parse_source_name(source, name):
    """Parse strings in form: <special_source>::<name> into (<special_source>, <name>)

    If :: is not present in the string, return (<source>, <name>)
    """
    if "::" in name:
        try:
            source, name = name.split("::")
        except ValueError:
            raise ValueError("The name: {0} couldn't be properly parsed. ".format(name) +
                             "Use the following synthax: <source>::<model/dataloader> or <model/dataloader>")

    # Check that the source is correctly specified
    available_sources = list(kipoi.config.model_sources().keys())
    if source not in available_sources:
        raise ValueError("Source: {0} is not present in the available sources: {1}".
                         format(source, available_sources))

    return (source, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


requirements = [
    "pyyaml",
    "future",
    "numpy",
    "pandas",
    "tqdm",
    "related>=0.6.0",
    "enum34",
    "colorlog",
    "jinja2",
    "cookiecutter",
    # sometimes required
    "h5py",
    "urllib3>=1.21.1,<1.23",
]

test_requirements = [
    "bumpversion",
    "wheel",
    "jedi",
    "epc",
    "pytest>=3.3.1",
    "pytest-xdist",  # running tests in parallel
    "pytest-pep8",  # see https://github.com/kipoi/kipoi/issues/91
    "pytest-cov",
    "coveralls",
    "scikit-learn",
    "cython",
    "keras",
    "tensorflow"
]

setup(
    name='kipoi',
    version='0.5.8',
    description="Kipoi: model zoo for genomics",
    author="Kipoi team",
    author_email='avsec@in.tum.de',
    url='https://github.com/kipoi/kipoi',
    long_description="Kipoi: model zoo for genomics. http://kipoi.org/",
    packages=find_packages(),
    install_requires=requirements,
    extras_require={
        "develop": test_requirements,
        "vep": ["kipoi_veff"],  # backcompatibily
    },
    entry_points={'console_scripts': ['kipoi = kipoi.__main__:main']},
    license="MIT license",
    zip_safe=False,
    keywords=["model zoo", "deep learning",
              "computational biology", "bioinformatics", "genomics"],
    test_suite='tests',
    include_package_data=True,
    tests_require=test_requirements
)



from __future__ import absolute_import

__author__ = 'Kipoi team'
__email__ = 'avsec@in.tum.de'
__version__ = '0.5.8'


# available modules
from . import config
from . import pipeline
from . import utils
from . import sources
from . import remote  # backward compat
from . import model
from . import data_utils
from . import data
from . import conda
from . import specs
from . import components  # backward compat
from . import readers
from . import writers
from . import plugin

# shortcuts
from .model import get_model
from .data import get_dataloader_factory
from .sources import get_model_descr, get_dataloader_descr
from .pipeline import install_model_requirements, install_dataloader_requirements
from .config import get_source, list_sources, list_models, list_dataloaders
from . import cli
from .plugin import list_plugins
# from .config import model_sources as sources
